#!/usr/bin/env python
import unittest

from ltpylib import collect


class TestCollect(unittest.TestCase):

  def test_to_csv(self):
    assert collect.to_csv(None) is None
    assert collect.to_csv(["a", "b"]) == "a,b"
    assert collect.to_csv(["a", "b"], sep=" | ") == "a | b"


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
import os
import subprocess
import sys
from pathlib import Path
from typing import IO, List, Optional, Sequence, Union

from ltpylib import files, filters, procs

FIND_REPOS_RECURSION_EXCLUDES = frozenset([
  'node_modules',
])


def create_git_cmd(
  git_args: Union[str, List[str]],
) -> List[str]:
  git_cmd = ["git"]
  if isinstance(git_args, str):
    git_cmd.append(git_args)
  else:
    git_cmd.extend(git_args)

  return git_cmd


def run_git_cmd(
  git_args: Union[str, List[str]],
  cwd: Union[Path, str] = os.getcwd(),
  check: bool = True,
  stderr: Optional[Union[int, IO]] = sys.stderr,
  log_cmd: bool = False,
) -> subprocess.CompletedProcess:
  return procs.run(create_git_cmd(git_args), check=check, cwd=cwd, stderr=stderr, log_cmd=log_cmd)


def run_git_cmd_stdout(
  git_args: Union[str, List[str]],
  cwd: Union[Path, str] = os.getcwd(),
  check: bool = True,
  stderr: Optional[Union[int, IO]] = sys.stderr,
  log_cmd: bool = False,
  strip: bool = True,
) -> str:
  result = run_git_cmd(git_args, cwd=cwd, check=check, stderr=stderr, log_cmd=log_cmd)

  if strip and result.stdout:
    return result.stdout.strip()

  return result.stdout


def run_git_cmd_regular_stdout(
  git_args: Union[str, List[str]],
  cwd: Union[Path, str] = os.getcwd(),
  check: bool = True,
  log_cmd: bool = False,
  **kwargs,
) -> subprocess.CompletedProcess:
  return procs.run_with_regular_stdout(create_git_cmd(git_args), cwd=cwd, check=check, log_cmd=log_cmd, **kwargs)


def base_dir(cwd: Union[Path, str] = os.getcwd()) -> Path:
  return Path(run_git_cmd_stdout("base-dir", cwd=cwd))


def current_branch(cwd: Union[Path, str] = os.getcwd()) -> str:
  return run_git_cmd_stdout("current-branch", cwd=cwd)


def default_branch(cwd: Union[Path, str] = os.getcwd()) -> str:
  return run_git_cmd_stdout("default-branch", cwd=cwd)


def repo_name(cwd: Union[Path, str] = os.getcwd()) -> str:
  return run_git_cmd_stdout("repo-name", cwd=cwd)


def repo_name_with_owner(cwd: Union[Path, str] = os.getcwd()) -> str:
  return run_git_cmd_stdout("repo-name-with-owner", cwd=cwd)


def repo_owner(cwd: Union[Path, str] = os.getcwd()) -> str:
  return run_git_cmd_stdout("repo-owner", cwd=cwd)


def in_base_dir(cwd: Union[Path, str] = os.getcwd()) -> bool:
  return run_git_cmd("in-base-dir", cwd=cwd).returncode == 0


def in_repo(cwd: Union[Path, str] = os.getcwd()) -> bool:
  return run_git_cmd("in-repo", cwd=cwd).returncode == 0


def diff_show(
  cwd: Union[Path, str] = os.getcwd(),
  diff_file: Union[Path, str] = None,
  diff_to_head: bool = True,
  log_cmd: bool = False,
) -> bool:
  git_args = ["--no-pager", "diff"]

  if diff_to_head:
    git_args.append("HEAD")

  if diff_file:
    diff_file_path = files.convert_to_path(diff_file)
    git_args.append(diff_file_path.relative_to(cwd).as_posix() if diff_file_path.is_absolute() else diff_file_path.as_posix())

  return run_git_cmd_regular_stdout(git_args, cwd=cwd, check=False, log_cmd=log_cmd).returncode == 1


def commit_show(
  message: str,
  cwd: Union[Path, str] = os.getcwd(),
  repo_files: Sequence[Union[Path, str]] = None,
  add_first: bool = True,
  check: bool = True,
  log_cmd: bool = False,
) -> subprocess.CompletedProcess:
  file_args = [files.convert_to_path(f).relative_to(cwd).as_posix() for f in repo_files if f] if repo_files else []
  if add_first:
    run_git_cmd_regular_stdout(["add"] + file_args, cwd=cwd, check=check, log_cmd=log_cmd)

  return run_git_cmd_regular_stdout(["commit", "-m", message] + file_args, cwd=cwd, check=check, log_cmd=log_cmd)


def is_file_part_of_git_repo(file_path: Path) -> bool:
  return git_repo_root_for_file(file_path) is not None


def git_repo_root_for_file(file_path: Path) -> Optional[Path]:
  if file_path.is_dir() and file_path.joinpath(".git").is_dir():
    return file_path

  for parent_file in file_path.parents:
    if parent_file.joinpath(".git").is_dir():
      return parent_file

  return None


def resolve_file_relative_to_git_base_dir(file_path: Path, current_dir: Path = Path(os.getcwd())) -> Optional[Path]:
  git_repo_root = git_repo_root_for_file(current_dir)
  if not git_repo_root:
    return None

  maybe_file = git_repo_root.joinpath(file_path)
  return maybe_file if maybe_file.exists() else None


def is_valid_repo(repo: Path) -> bool:
  if not repo.is_dir():
    return False

  if repo.joinpath('.git').is_dir():
    return True

  return in_repo(repo)


def filter_invalid_repos(git_repos: List[Path]) -> List[Path]:
  filtered = []
  for repo in git_repos:
    if not is_valid_repo(repo):
      continue

    filtered.append(repo)

  return filtered


def print_repos(git_repos: List[Path]):
  for repo in git_repos:
    print(repo.as_posix())


def find_git_repos(
  base_dir: Path,
  max_depth: int = -1,
  recursion_include_patterns: Sequence[str] = None,
  recursion_exclude_patterns: Sequence[str] = None,
  recursion_includes: Sequence[str] = None,
  recursion_excludes: Sequence[str] = FIND_REPOS_RECURSION_EXCLUDES
) -> List[Path]:
  dotgit_dirs = files.find_children(
    base_dir,
    break_after_match=True,
    include_files=False,
    max_depth=max_depth,
    includes=['.git'],
    recursion_include_patterns=recursion_include_patterns,
    recursion_exclude_patterns=recursion_exclude_patterns,
    recursion_includes=recursion_includes,
    recursion_excludes=recursion_excludes
  )
  return [dotgit.parent for dotgit in dotgit_dirs]


def add_git_dirs(
  git_repos: List[Path],
  add_dir: List[Path],
  include_patterns: List[str] = None,
  exclude_patterns: List[str] = None,
  max_depth: int = -1,
  recursion_include_patterns: Sequence[str] = None,
  recursion_exclude_patterns: Sequence[str] = None,
  recursion_includes: Sequence[str] = None,
  recursion_excludes: Sequence[str] = FIND_REPOS_RECURSION_EXCLUDES
) -> List[Path]:
  for git_dir in add_dir:
    if not git_dir.is_dir():
      continue

    add_dir_repos = find_git_repos(
      git_dir,
      max_depth=max_depth,
      recursion_include_patterns=recursion_include_patterns,
      recursion_exclude_patterns=recursion_exclude_patterns,
      recursion_includes=recursion_includes,
      recursion_excludes=recursion_excludes
    )
    add_dir_repos.sort()
    for git_repo in add_dir_repos:
      if filters.should_skip(git_repo, exclude_patterns=exclude_patterns, include_patterns=include_patterns):
        continue

      if git_repo in git_repos:
        continue

      git_repos.append(git_repo)

  return git_repos



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
import re
from typing import Dict, List, Tuple, Union

import jira.resources
from jira import JIRA
from requests import Session

from ltpylib import inputs, strconverters, strings
from ltpylib.collect import EMPTY_LIST, EMPTY_MAP, to_csv
from ltpylib.jira_api_types import Issue, IssueSearchResult, JiraProject, Sprint, SprintReport, SprintState, VelocityReport

OPTION_AGILE_REST_PATH = "agile_rest_path"

JIRA_API_SEARCH: str = "/rest/api/2/search"
JIRA_API_EPICS: str = "/rest/greenhopper/latest/xboard/plan/backlog/epics"
JIRA_API_SPRINT_REPORT: str = "/rest/greenhopper/1.0/rapid/charts/sprintreport"
JIRA_API_VELOCITY_REPORT: str = "/rest/greenhopper/1.0/rapid/charts/velocity.json"

ISSUE_FIELD_SPRINT_FINAL: str = "sprintFinal"
ISSUE_FIELD_SPRINT_RAW: str = "sprintRaw"


class JiraApi(object):

  def __init__(
    self,
    api: JIRA = None,
    url: str = None,
    auth: Tuple[str, str] = None,
    basic_auth: Tuple[str, str] = None,
  ):
    if api is not None:
      self.api: JIRA = api
    elif url is not None and (auth is not None or basic_auth is not None):
      self.api: JIRA = JIRA(
        url,
        auth=auth,
        basic_auth=basic_auth,
        options={
          OPTION_AGILE_REST_PATH: jira.resources.AgileResource.AGILE_BASE_REST_PATH,
        },
      )
    else:
      raise Exception("Must be initialized with 'api: JIRA' instance or both 'url' and 'auth'")

  def get_session(self) -> Session:
    return self.api._session

  def board_id(
    self,
    board_id_or_name: Union[str, int],
  ) -> int:
    if isinstance(board_id_or_name, int) or board_id_or_name.isdigit():
      return int(board_id_or_name)

    boards = self.api.boards(maxResults=0)
    for maybe_board in boards:
      if maybe_board.id == board_id_or_name or maybe_board.name == board_id_or_name:
        return int(maybe_board.id)

    raise Exception("Could not find board=%s, choices below:\n%s" % (board_id_or_name, "\n".join(["%s (%s)" % (board.id, board.name) for board in boards])))

  def epics(
    self,
    board_id: int,
  ) -> dict:
    return self.get_session().get(self.api.client_info() + JIRA_API_EPICS + "?rapidViewId=" + str(board_id)).json()

  def issue(
    self,
    id: str,
    fields: List[str] = None,
    expand: List[str] = None,
    # parse response config
    no_convert: bool = False,
    convert_single_value_arrays: bool = False,
    create_new_result: bool = False,
    skip_fields: List[str] = EMPTY_LIST,
    dict_field_to_inner_field: Dict[str, str] = EMPTY_MAP,
    join_array_fields: List[str] = EMPTY_LIST,
    date_fields: List[str] = EMPTY_LIST
  ) -> Issue:
    return Issue(
      values=JiraApi.parse_api_response_with_names(
        self.api.issue(
          id,
          fields=to_csv(fields),
          expand=JiraApi.expand_with_names(expand),
        ).raw,
        no_convert=no_convert,
        convert_single_value_arrays=convert_single_value_arrays,
        create_new_result=create_new_result,
        skip_fields=skip_fields,
        dict_field_to_inner_field=dict_field_to_inner_field,
        join_array_fields=join_array_fields,
        date_fields=date_fields
      )
    )

  def issue_summaries(
    self,
    issues: List[str],
    markdown: bool = False,
    link: bool = False,
  ) -> List[str]:
    summaries = []

    for issue in issues:
      if issue.count("/") > 0:
        issue_key = issue.split("/")[-1]
      else:
        issue_key = issue

      result = self.issue(issue_key)

      if markdown:
        summary: str = "[%s](%s/browse/%s) `%s`" % (self.url, result.key, result.key, result.summary)
      elif link:
        summary: str = "%s/browse/%s %s" % (self.url, result.key, result.summary)
      else:
        summary: str = "%s %s" % (result.key, result.summary)

      summaries.append(summary)

    return summaries

  def project(
    self,
    project_id_or_name: Union[str, int],
  ) -> JiraProject:
    jira_project: jira.resources.Project = self.api.project(project_id_or_name)
    return JiraProject(values=jira_project.raw)

  def search_issues(
    self,
    jql_or_filter_id: Union[str, int],
    start_at: int = 0,
    max_results: Union[int, bool] = 50,
    validate_query: bool = True,
    fields: List[str] = None,
    expand: List[str] = None,
    json_result: bool = True,
    # parse response config
    no_convert: bool = False,
    convert_single_value_arrays: bool = False,
    create_new_result: bool = False,
    skip_fields: List[str] = EMPTY_LIST,
    dict_field_to_inner_field: Dict[str, str] = EMPTY_MAP,
    join_array_fields: List[str] = EMPTY_LIST,
    date_fields: List[str] = EMPTY_LIST,
  ) -> IssueSearchResult:
    if isinstance(jql_or_filter_id, int) or jql_or_filter_id.isdigit():
      jira_filter: jira.resources.Filter = self.api.filter(jql_or_filter_id)
      jql: str = jira_filter.jql
    else:
      jql: str = jql_or_filter_id

    get_all = isinstance(max_results, bool) and not max_results
    if json_result:
      check_for_more = get_all or max_results > 100
      max_results_param = 100 if check_for_more else max_results
    else:
      check_for_more = False
      max_results_param = max_results

    fields_csv = to_csv(fields)
    expanded_with_names = JiraApi.expand_with_names(expand)
    result = IssueSearchResult(
      values=JiraApi.parse_api_response_with_names(
        self.api.search_issues(
          jql,
          startAt=start_at,
          maxResults=max_results_param,
          validate_query=validate_query,
          fields=fields_csv,
          expand=expanded_with_names,
          json_result=json_result,
        ),
        "issues",
        no_convert=no_convert,
        convert_single_value_arrays=convert_single_value_arrays,
        create_new_result=create_new_result,
        skip_fields=skip_fields,
        dict_field_to_inner_field=dict_field_to_inner_field,
        join_array_fields=join_array_fields,
        date_fields=date_fields
      )
    )

    if check_for_more:
      last_result = result

      while last_result.issues and len(result.issues) < result.total and (get_all or len(result.issues) < max_results):
        more_start_at = last_result.startAt + last_result.maxResults

        last_result = IssueSearchResult(
          values=JiraApi.parse_api_response_with_names(
            self.api.search_issues(
              jql,
              startAt=more_start_at,
              maxResults=max_results_param,
              validate_query=validate_query,
              fields=fields_csv,
              expand=expanded_with_names,
              json_result=json_result,
            ),
            "issues",
            no_convert=no_convert,
            convert_single_value_arrays=convert_single_value_arrays,
            create_new_result=create_new_result,
            skip_fields=skip_fields,
            dict_field_to_inner_field=dict_field_to_inner_field,
            join_array_fields=join_array_fields,
            date_fields=date_fields
          )
        )

        result.issues.extend(last_result.issues)

    return result

  def sprint_id(
    self,
    sprint_id_or_name: Union[str, int],
    board_id: int,
    allow_select_sprint: bool = True,
    find_active_sprint: bool = False,
    find_first_future_sprint: bool = False,
    include_closed_sprints: bool = False,
  ) -> int:
    if not sprint_id_or_name:
      if not allow_select_sprint:
        raise ValueError("sprint_id_or_name is None and allow_select_sprint=False")

      return self.sprint_selection(
        board_id,
        find_active_sprint=find_active_sprint,
        find_first_future_sprint=find_first_future_sprint,
        include_closed_sprints=include_closed_sprints,
      )

    if isinstance(sprint_id_or_name, int) or sprint_id_or_name.isdigit():
      return int(sprint_id_or_name)

    sprints = self.api.sprints(board_id)
    for maybe_sprint in sprints:
      if maybe_sprint.id == sprint_id_or_name or maybe_sprint.name == sprint_id_or_name:
        return int(maybe_sprint.id)

    raise Exception("Could not find sprint=%s, choices below:\n%s" % (sprint_id_or_name, "\n".join(JiraApi.create_sprint_choices(sprints))))

  def sprint_issues(
    self,
    board_id_or_name: Union[str, int],
    sprint_id_or_name: Union[str, int] = None,
    find_active_sprint: bool = False,
    find_first_future_sprint: bool = False,
    include_closed_sprints: bool = False,
    add_filters: List[str] = None,
    include_placeholder: bool = False,
    include_sub_tasks: bool = False,
    unassigned: bool = False,
    unestimated: bool = False,
    fields: List[str] = None,
    max_results: int = 500,
  ) -> IssueSearchResult:
    board_id = self.board_id(board_id_or_name)
    sprint_id = self.sprint_id(
      sprint_id_or_name,
      board_id,
      allow_select_sprint=True,
      find_active_sprint=find_active_sprint,
      find_first_future_sprint=find_first_future_sprint,
      include_closed_sprints=include_closed_sprints,
    )

    add_filters = " ".join(
      JiraApi.create_filters(
        add_filters=add_filters,
        include_placeholder=include_placeholder,
        include_sub_tasks=include_sub_tasks,
        unassigned=unassigned,
        unestimated=unestimated,
      )
    )

    jql = "Sprint = %s %s ORDER BY updated DESC" % (sprint_id, add_filters)

    return self.search_issues(jql, max_results=max_results, fields=fields)

  def sprint_report(
    self,
    board_id: int,
    sprint_id: int,
  ) -> SprintReport:
    url = self.api.client_info() + JIRA_API_SPRINT_REPORT + "?rapidViewId=%s&sprintId=%s" % (board_id, sprint_id)
    return SprintReport(self.get_session().get(url).json())

  def sprint_selection(
    self,
    board_id: int,
    find_active_sprint: bool = False,
    find_first_future_sprint: bool = False,
    include_closed_sprints: bool = False,
    include_closed_sprints_count: int = 2,
  ) -> int:
    sprints = self.sprints(
      board_id,
      include_closed_sprints=include_closed_sprints,
      include_closed_sprints_count=include_closed_sprints_count,
    )
    choices = JiraApi.create_sprint_choices(sprints)

    if find_active_sprint:
      for sprint in sprints:
        if sprint.state == "ACTIVE":
          return int(sprint.id)

      raise Exception("No ACTIVE sprint found in sprints below:\n%s" % "\n".join(choices))
    elif find_first_future_sprint:
      for sprint in sprints:
        if sprint.state == "FUTURE":
          return int(sprint.id)

      raise Exception("No FUTURE sprint found in sprints below:\n%s" % "\n".join(choices))

    choice = inputs.select_prompt(choices)
    return int(strings.substring_before(choice, " ("))

  def sprints(
    self,
    board_id: int,
    include_closed_sprints: bool = True,
    include_closed_sprints_count: int = None,
  ) -> List[Sprint]:
    if self.using_greenhopper():
      state = None
    else:
      sprint_states = ["active", "future"]
      if include_closed_sprints:
        sprint_states.append("closed")
      state = ",".join(sprint_states)

    sprints: List[Sprint] = [Sprint(sprint.raw) for sprint in self.api.sprints(board_id, state=state)]
    if not include_closed_sprints:
      sprints = [sprint for sprint in sprints if sprint.state != SprintState.CLOSED]
    elif include_closed_sprints_count is not None and include_closed_sprints_count > 0:
      closed_sprints: List[Sprint] = [sprint for sprint in sprints if sprint.state == SprintState.CLOSED]
      other_sprints: List[Sprint] = [sprint for sprint in sprints if sprint.state != SprintState.CLOSED]
      closed_sprints = sorted(closed_sprints, key=lambda spr: spr.endDate, reverse=True)[:include_closed_sprints_count]
      sprints = sorted(other_sprints + closed_sprints, key=lambda spr: spr.id)

    return sprints

  def using_greenhopper(self) -> bool:
    return self.api._options[OPTION_AGILE_REST_PATH] == jira.resources.GreenHopperResource.GREENHOPPER_REST_PATH

  def velocity_report(
    self,
    board_id: int,
  ) -> VelocityReport:
    url = self.api.client_info() + JIRA_API_VELOCITY_REPORT + "?rapidViewId=%s" % (board_id)
    return VelocityReport(self.get_session().get(url).json())

  @staticmethod
  def create_filters(
    add_filters: List[str] = None,
    include_placeholder: bool = False,
    include_sub_tasks: bool = False,
    unassigned: bool = False,
    unestimated: bool = False,
  ) -> List[str]:
    filters: List[str] = []
    if add_filters:
      filters.extend(add_filters)

    if not include_placeholder:
      filters.append("AND (labels IS EMPTY OR labels NOT IN (placeholder))")

    if not include_sub_tasks:
      filters.append("AND type NOT IN (Sub-task)")

    if unassigned:
      filters.append("AND assignee IS NULL")

    if unestimated:
      filters.append("AND \"Story Points\" IS NULL")

    return filters

  @staticmethod
  def create_issues_display(issues: List[Issue], fields: List[str]) -> List[str]:
    return sorted([" ".join([getattr(issue, field) for field in fields]) for issue in issues])

  @staticmethod
  def create_sprint_choices(sprints: List[Sprint]) -> List[str]:
    return [('%s (%s): %s' % (sprint.id, sprint.state.name, sprint.name)) for sprint in sprints]

  @staticmethod
  def expand_with_names(expand: List[str]) -> str:
    if expand is None:
      expand = []

    expand.append("names")
    return to_csv(expand)

  @staticmethod
  def parse_api_response_with_names(
    result: dict,
    convert_values_field: str = None,
    names_field: str = "names",
    fields_field: str = "fields",
    no_convert: bool = False,
    convert_single_value_arrays: bool = False,
    create_new_result: bool = False,
    skip_fields: List[str] = EMPTY_LIST,
    dict_field_to_inner_field: Dict[str, str] = EMPTY_MAP,
    join_array_fields: List[str] = EMPTY_LIST,
    date_fields: List[str] = EMPTY_LIST
  ) -> dict:
    if not convert_values_field:
      convert_values: List[Dict] = [result]
    else:
      convert_values_temp: Union[dict, List[dict]] = result.get(convert_values_field)
      if isinstance(convert_values_temp, list):
        convert_values: List[Dict] = convert_values_temp
      else:
        convert_values: List[Dict] = [convert_values_temp]

    names: Dict[str, str] = result.get(names_field)
    array_fields: List[str] = []
    convert_array_fields: List[str] = []

    for convert_value in convert_values:
      if create_new_result:
        updated_value: Dict = {}
      else:
        updated_value: Dict = convert_value

      if create_new_result:
        for entry in convert_value.items():
          if entry[0] != fields_field and not entry[0] in skip_fields:
            updated_value[entry[0]] = entry[1]
      elif skip_fields:
        for field in skip_fields:
          if field != fields_field and field in updated_value:
            updated_value.pop(field)

      if create_new_result:
        issue_fields: Dict = convert_value.get(fields_field, {})
      else:
        issue_fields: Dict = convert_value.pop(fields_field, {})

      for entry in issue_fields.items():
        key: str = entry[0]
        val = entry[1]
        if val is None:
          continue
        elif isinstance(val, list) and not val:
          continue
        elif no_convert:
          updated_value[key] = val
          continue

        orig_key: str = key
        if key.startswith("customfield_") and key in names:
          key = strconverters.to_camel_case(names.get(key))

        if key in skip_fields or orig_key in skip_fields:
          continue
        elif key == "sprint":
          sprints: List[Union[str, dict]] = val
          if isinstance(sprints[0], str):
            val = [re.match(r".*,name=(.*?),.*", sprint).group(1) for sprint in sprints]
          else:
            val = [sprint.get("name") for sprint in sprints]

          if ISSUE_FIELD_SPRINT_RAW not in skip_fields:
            updated_value[ISSUE_FIELD_SPRINT_RAW] = sprints
          if ISSUE_FIELD_SPRINT_FINAL not in skip_fields:
            updated_value[ISSUE_FIELD_SPRINT_FINAL] = val[-1]
        elif key in dict_field_to_inner_field:
          if key not in skip_fields:
            updated_value[key] = val
          inner_field: str = dict_field_to_inner_field.get(key)
          if isinstance(val, dict):
            val = val.get(inner_field)
            key = key + "_" + inner_field
          elif isinstance(val, list):
            val = [elem.get(inner_field) for elem in val]
            key = key + "_" + inner_field

        if isinstance(val, list):
          if key in join_array_fields:
            val = ",".join(val)
          elif convert_single_value_arrays:
            if key not in array_fields:
              array_fields.append(key)
              convert_array_fields.append(key)

            if len(val) > 1:
              if key in convert_array_fields:
                convert_array_fields.remove(key)
        # elif key in date_fields:
        #   val = val.replace("-0400", "").replace("-0500", "").replace("T", " ")

        updated_value[key] = val

    if not no_convert and convert_single_value_arrays and convert_array_fields:
      for field in convert_array_fields:
        for convert_value in convert_values:
          if field in convert_value and convert_value.get(field):
            convert_value[field] = "\n".join(convert_value.get(field))

    return result

  @property
  def url(self):
    return self.api._options["server"]



#!/usr/bin/env python
import dataclasses
from typing import Any, List, Optional


@dataclasses.dataclass
class SearchParam:
  name: Optional[str]
  value: Any
  is_or: bool = False


def create_search_query(params: List[SearchParam]) -> str:
  query: List[str] = []

  for param in params:
    query.append(construct_search_param(param))

  return " ".join(query)


def construct_search_param(param: SearchParam) -> str:
  part = f"{param.name}:" if param.name else ""
  value = param.value
  if isinstance(value, list) and len(value) == 1:
    value = value[0]

  if isinstance(value, list):
    part += "("

    if len(value) > 0 and isinstance(value[0], SearchParam):
      values = [construct_search_param(v) for v in value]
    else:
      values = [str(v) for v in value]

    if param.is_or:
      part += " OR ".join(values)
    else:
      part += " AND ".join(values)

    part += ")"
  else:
    part += construct_search_param(value) if isinstance(value, SearchParam) else str(value)

  return part



#!/usr/bin/env python
import json
import os
import tempfile
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

from ltpylib.collect import modify_list_of_dicts
from ltpylib.common_types import TypeWithDictRepr
from ltpylib.dicts import convert_boolean_values_to_string, modify_dict_fields

CUSTOM_JSON_DUMPERS: Dict[str, Tuple[Callable[[Any], Any], Optional[Callable[[Any], bool]]]] = {}

HAS_JQ: Optional[bool] = None

PYGMENTS_DEFAULT_STYLE_ORDER = ["jq", "smyck", "vim", "solarized-light"]
PYGMENTS_DEFAULT_STYLE: Optional[str] = None


def default_pygments_style() -> str:
  global PYGMENTS_DEFAULT_STYLE

  if PYGMENTS_DEFAULT_STYLE:
    return PYGMENTS_DEFAULT_STYLE

  from pygments.plugin import iter_entry_points, STYLE_ENTRY_POINT

  for desired_style in PYGMENTS_DEFAULT_STYLE_ORDER:
    for entrypoint in iter_entry_points(STYLE_ENTRY_POINT):
      if entrypoint.name == desired_style:
        PYGMENTS_DEFAULT_STYLE = desired_style
        return PYGMENTS_DEFAULT_STYLE

  raise ValueError("Could not find pygments style from options: %s" % PYGMENTS_DEFAULT_STYLE_ORDER)


def find_pygments_style() -> str:
  pygments_style = os.getenv("PYGMENTS_STYLE")
  if not pygments_style:
    pygments_style = default_pygments_style()

  return pygments_style


def create_terminal_formatter(pygments_style: str):
  if "256" in os.getenv("TERM", ""):
    from pygments.formatters.terminal256 import Terminal256Formatter

    return Terminal256Formatter(style=pygments_style)

  from pygments.formatters.terminal import TerminalFormatter

  return TerminalFormatter(style=pygments_style)


def has_jq() -> bool:
  global HAS_JQ
  if HAS_JQ is not None:
    return HAS_JQ

  from ltpylib.checks import check_command

  HAS_JQ = check_command("jq")
  return HAS_JQ


def colorize_json_jq(data: Union[str, dict, Sequence]) -> Optional[str]:
  from ltpylib.procs import run_and_parse_output

  if data is None:
    json_data = "null"
  else:
    json_data = data if isinstance(data, str) or data is None else json.dumps(data, indent=None, default=json_dump_default)

  return run_and_parse_output(["jq", "--sort-keys", "--color-output"], input=json_data, check=True)[1]


def colorize_with_pygments(format_lang: str, prettifier: Callable, data: Union[str, dict, Sequence], pygments_style: str = None) -> Union[bytes, str]:
  from pygments import highlight
  from pygments.lexers import get_lexer_by_name

  if not pygments_style:
    pygments_style = find_pygments_style()

  if data is None:
    string_data = "null"
  elif isinstance(data, str):
    string_data = data
  else:
    string_data = prettifier(data, colorize=False)

  return highlight(
    string_data,
    get_lexer_by_name(format_lang),
    create_terminal_formatter(pygments_style),
  )


def colorize_json(data: Union[str, dict, Sequence], pygments_style: str = None, force_pygments: bool = False) -> Union[bytes, str]:
  global HAS_JQ
  if not force_pygments and pygments_style is None and HAS_JQ is not False:
    if HAS_JQ is None:
      try:
        result = colorize_json_jq(data)
        HAS_JQ = True
        return result
      except FileNotFoundError:
        HAS_JQ = False

    else:
      return colorize_json_jq(data)

  return colorize_with_pygments("json", prettify_json, data, pygments_style=pygments_style)


def colorize_xml(data: Union[str, dict, Sequence], pygments_style: str = None) -> Union[bytes, str]:
  return colorize_with_pygments("xml", prettify_xml, data, pygments_style=pygments_style)


def colorize_yaml(data: Union[str, dict, Sequence], pygments_style: str = None) -> Union[bytes, str]:
  return colorize_with_pygments("yaml", prettify_yaml, data, pygments_style=pygments_style)


def is_output_to_terminal() -> bool:
  import sys

  return sys.stdout.isatty()


def should_color(colorize: bool = False, auto_color: bool = False) -> bool:
  if colorize:
    return True

  if auto_color and is_output_to_terminal():
    return True

  return False


def add_custom_json_dumper(dumper_id: str, dumper: Callable[[Any], Any], use_if: Callable[[Any], bool] = None):
  CUSTOM_JSON_DUMPERS[dumper_id] = (dumper, use_if)


def json_dump_default(val: Any) -> Any:
  if hasattr(val, "to_dict"):
    return getattr(val, "to_dict")()

  if CUSTOM_JSON_DUMPERS:
    for dumper, use_if in CUSTOM_JSON_DUMPERS.values():
      if use_if is not None:
        if not use_if(val):
          continue
        return dumper(val)
      else:
        dumper_val = dumper(val)
        if dumper_val is not None:
          return dumper_val

  return getattr(val, '__dict__', str(val))


def load_json_remove_nulls(data: str) -> Any:
  from ltpylib import dicts

  return json.loads(
    data,
    object_hook=dicts.remove_nulls_and_empty,
  )


def prettify_json_compact(
  obj,
  remove_nulls: bool = False,
  colorize: bool = False,
  auto_color: bool = False,
) -> str:
  return prettify_json(obj, remove_nulls=remove_nulls, colorize=colorize, auto_color=auto_color, compact=True)


def prettify_json_auto_color(
  obj,
  remove_nulls: bool = False,
  compact: bool = False,
) -> str:
  return prettify_json(obj, remove_nulls=remove_nulls, auto_color=True, compact=compact)


def prettify_json(
  obj,
  remove_nulls: bool = False,
  colorize: bool = False,
  auto_color: bool = False,
  compact: bool = False,
) -> str:
  if remove_nulls:
    obj = load_json_remove_nulls(json.dumps(obj, default=json_dump_default))

  output = json.dumps(
    obj,
    sort_keys=True,
    indent=None if compact else '  ',
    separators=(",", ":") if compact else None,
    default=json_dump_default,
  )

  if should_color(colorize=colorize, auto_color=auto_color):
    output = colorize_json(output)

  return output


def prettify_json_remove_nulls(obj) -> str:
  return prettify_json(obj, remove_nulls=True)


def prettify_xml(obj, remove_nulls: bool = False, colorize: bool = False, auto_color: bool = False) -> str:
  from xml.dom.minidom import parseString
  from dicttoxml import dicttoxml

  if remove_nulls:
    obj = load_json_remove_nulls(json.dumps(obj, default=json_dump_default))

  output = parseString(dicttoxml(obj)).toprettyxml()

  if should_color(colorize=colorize, auto_color=auto_color):
    output = colorize_xml(output)

  return output


def prettify_yaml(obj, remove_nulls: bool = False, colorize: bool = False, auto_color: bool = False, sort_keys: bool = True) -> str:
  import yaml

  if remove_nulls:
    obj = load_json_remove_nulls(json.dumps(obj, default=json_dump_default))

  output = yaml.dump(
    obj,
    default_flow_style=False,
    sort_keys=sort_keys,
  )

  if should_color(colorize=colorize, auto_color=auto_color):
    output = colorize_yaml(output)

  return output


def prettify_sql(sql: str) -> str:
  from ltpylib.files import read_file, write_file
  from ltpylib.files_prettifier import prettify_sql_file

  tmp_sql_file = Path(tempfile.mktemp(suffix=".sql"))
  try:
    write_file(tmp_sql_file, sql)
    prettify_sql_file(tmp_sql_file)
    return read_file(tmp_sql_file)
  finally:
    tmp_sql_file.unlink(missing_ok=True)


def dicts_to_csv(
  data: Union[List[dict], List[TypeWithDictRepr]],
  showindex: bool = False,
  sep: str = ",",
  header: bool = True,
  fields_included: Sequence[str] = None,
  fields_order: Sequence[str] = None,
  fields_order_from_included: bool = False,
  modify_in_place: bool = False,
  convert_booleans_to_string: bool = False,
) -> str:
  from pandas import DataFrame

  if fields_included and not fields_order and fields_order_from_included:
    fields_order = fields_included

  data_as_dicts: List[dict] = create_data_as_dicts(
    data,
    fields_included=fields_included,
    fields_order=fields_order,
    modify_in_place=modify_in_place,
    convert_booleans_to_string=convert_booleans_to_string,
  )

  data_frame = DataFrame(data_as_dicts)
  return data_frame.to_csv(
    sep=sep,
    header=header,
    index=showindex,
  )


def dicts_to_markdown_table(
  data: Union[List[dict], List[TypeWithDictRepr]],
  showindex: bool = False,
  tablefmt: str = "github",
  escape_data: bool = True,
  headers: Sequence[str] = None,
  fields_included: Sequence[str] = None,
  fields_order: Sequence[str] = None,
  fields_order_from_included: bool = False,
  modify_in_place: bool = False,
  convert_booleans_to_string: bool = False,
) -> str:
  import tabulate

  from pandas import DataFrame

  if fields_included and not fields_order and fields_order_from_included:
    fields_order = fields_included

  data_as_dicts: List[dict] = create_data_as_dicts(
    data,
    fields_included=fields_included,
    fields_order=fields_order,
    modify_in_place=modify_in_place,
    convert_booleans_to_string=convert_booleans_to_string,
  )

  if escape_data:

    def escape_fn(row: dict):
      for field, value in row.items():
        if isinstance(value, str) and ("|" in value or "\n" in value):
          row[field] = value.replace("|", "&#124;").replace("\n", "<br/>")

    data_as_dicts = modify_list_of_dicts(data_as_dicts, escape_fn, in_place=modify_in_place)

  data_frame = DataFrame(data_as_dicts)
  return tabulate.tabulate(
    data_frame,
    showindex=showindex,
    headers=headers if headers is not None and len(headers) > 0 else data_frame.columns,
    tablefmt=tablefmt,
  )


def sort_csv_rows(rows: List[str]) -> List[str]:
  return [rows[0]] + sorted(rows[1:])


def create_data_as_dicts(
  data: Union[List[dict], List[TypeWithDictRepr]],
  fields_included: Sequence[str] = None,
  fields_order: Sequence[str] = None,
  modify_in_place: bool = False,
  convert_booleans_to_string: bool = False,
) -> List[dict]:
  data_as_dicts: List[dict] = data

  if len(data) > 0 and isinstance(data[0], TypeWithDictRepr):
    data_as_class: List[TypeWithDictRepr] = data
    data_as_dicts = [val.as_dict() for val in data_as_class]

  if fields_order or fields_included:
    data_as_dicts = modify_dict_fields(
      data_as_dicts,
      fields_included=fields_included,
      fields_order=fields_order,
      in_place=modify_in_place,
    )

  if convert_booleans_to_string:
    data_as_dicts = convert_boolean_values_to_string(data_as_dicts, only_fields=fields_included)

  return data_as_dicts



#!/usr/bin/env python

import os
from typing import Optional


class PutEnvTemp(object):

  def __init__(self, var_name: str, var_value: str):
    self.var_name = var_name
    self.var_value = var_value
    self.reset_value: Optional[str] = None

  def update_env(self, to_val: Optional[str]):
    if to_val is None:
      os.unsetenv(self.var_name)
    else:
      os.putenv(self.var_name, to_val)

  def __enter__(self):
    self.reset_value = os.getenv(self.var_name)
    self.update_env(self.var_value)

  def __exit__(self, *exc_details):
    self.update_env(self.reset_value)


def putenv_temp(var_name: str, var_value: str) -> PutEnvTemp:
  return PutEnvTemp(var_name, var_value)



#!/usr/bin/env python
# pylint: disable=C0111

from typing import Callable, Dict, List, Optional, Union

from bs4 import BeautifulSoup

StrConverter = Callable[[str], str]


def create_parser(html: str) -> BeautifulSoup:
  return BeautifulSoup(html, 'html5lib')


def extract_text_from_html(html: str, selector: str) -> Union[str, None]:
  soup = create_parser(html)
  selected: BeautifulSoup = soup.select_one(selector)
  if not selected:
    return

  return selected.get_text()


def header_replacements_as_header_converter(header_replacements: Dict[str, str]) -> StrConverter:
  return lambda val: header_replacements.get(val, val)


def parse_table(
  html: str,
  table_selector: str,
  header_replacements: Dict[str, str] = None,
  header_converters: List[StrConverter] = None,
  skip_headers: List[str] = None,
  val_converters: List[StrConverter] = None,
  row_data_predicate: Callable[[dict], bool] = None
) -> Optional[List[dict]]:
  soup = create_parser(html)
  table: BeautifulSoup = soup.select_one(table_selector)
  if not table:
    return None

  table_headers = _find_table_headers(table, header_replacements, header_converters)
  rows = _find_table_rows(table)
  results: List[dict] = []
  for row in rows:
    if not table_headers:
      table_headers = _parse_header_row(row, header_replacements, header_converters)
      continue

    table_data = row.find_all('td')
    if table_data:
      row_data = {}
      for idx, data in enumerate(table_data):
        header = table_headers.get(idx)
        if not header:
          continue
        elif skip_headers and header in skip_headers:
          continue

        val: str = data.get_text().strip()
        if val_converters:
          for val_converter in val_converters:
            val = val_converter(val)

        if val:
          row_data[header] = val

      if row_data_predicate and not row_data_predicate(row_data):
        continue

      results.append(row_data)

  return results


def _find_table_headers(table: BeautifulSoup, header_replacements: Dict[str, str], header_converters: List[StrConverter]) -> Dict[int, str]:
  header_row = table.select_one('thead > tr')
  if not header_row:
    return None

  return _parse_header_row(header_row, header_replacements, header_converters)


def _find_table_rows(table: BeautifulSoup) -> List[BeautifulSoup]:
  tbody: BeautifulSoup = table.select_one('tbody')
  if tbody:
    return tbody.find_all('tr')

  return table.find_all('tr')


def _parse_header_row(header_row: BeautifulSoup, header_replacements: Dict[str, str], header_converters: List[StrConverter]) -> Dict[int, str]:
  header_cols: BeautifulSoup = header_row.select('td, th')
  if not header_cols:
    return None

  table_headers = {}
  found_headers = []
  for idx, header_elem in enumerate(header_cols):
    header = header_elem.get_text().strip()

    if not header:
      continue

    if header in found_headers:
      header_vers = 2
      new_header = header + str(header_vers)
      while new_header in found_headers:
        header_vers += 1
        new_header = header + str(header_vers)
        if header_vers >= 100:
          raise Exception("Too many duplicate headers: header=%s new_header=%s header_vers=%s" % (header, new_header, header_vers))

      header = new_header

    found_headers.append(header)

    if header_replacements:
      header = header_replacements.get(header, header)

    if header_converters:
      for header_converter in header_converters:
        header = header_converter(header)

      if not header:
        continue

    table_headers[idx] = header

  return table_headers


def _main():
  import sys

  method_params = []
  for arg in sys.argv[2:]:
    if arg == '--':
      method_params.append(sys.stdin.read())
    else:
      method_params.append(arg)

  result = globals()[sys.argv[1]](*method_params)
  if result is not None:
    print(result)


if __name__ == "__main__":
  try:
    _main()
  except KeyboardInterrupt:
    exit(130)



#!/usr/bin/env python
from pathlib import Path
from typing import List, Union

import yaml


class TestConfigInput(object):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.expected: str = values.pop("expected", None)
    self.input: str = values.pop("input", None)


class TestConfig(object):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.test_cases: List[TestConfigInput] = list(map(TestConfigInput, values.pop("test_cases", []))) if "test_cases" in values else None


def load_yaml_test_config_file(test_file: Union[str, bytes, Path]) -> TestConfig:
  with open(test_file, "r") as tf:
    config: TestConfig = TestConfig(values=yaml.full_load(tf))

  return config



#!/usr/bin/env python
import subprocess
from pathlib import Path
from typing import List, Optional, Tuple, Union

import requests

from ltpylib import procs, xmlparser
from ltpylib.collect import remove_nulls
from ltpylib.common_types import TypeWithDictRepr
from ltpylib.inputs import select_prompt
from ltpylib.requests_helper import maybe_throw

MVN_EXPR_PROJECT_GROUP_ID = 'project.groupId'
MVN_EXPR_PROJECT_ARTIFACT_ID = 'project.artifactId'
MVN_EXPR_PROJECT_VERSION = 'project.version'

DEFAULT_SEARCH_ROWS = 50


class MavenArtifact(TypeWithDictRepr):

  def __init__(
    self,
    group_id: str = None,
    artifact_id: str = None,
    version: str = None,
    packaging: str = None,
    classifier: str = None,
    scope: str = None,
  ):
    self.group_id: str = group_id
    self.artifact_id: str = artifact_id
    self.version: str = version
    self.packaging: str = packaging
    self.classifier: str = classifier
    self.scope: str = scope

  def __eq__(self, other):
    return self.__dict__ == other.__dict__

  def __hash__(self):
    return hash(str(self.__dict__))

  def to_artifact_string(self) -> str:
    # groupId:artifactId[:version[:packaging[:classifier[:scope]]]]
    parts_reversed: List[str] = []

    parts_ordered = [self.group_id, self.artifact_id, self.version, self.packaging, self.classifier, self.scope]
    has_part = False
    for part in reversed(parts_ordered):
      if part:
        has_part = True
        parts_reversed.append(part)
      elif has_part:
        parts_reversed.append("*")

    return ":".join(reversed(parts_reversed))

  def to_dict(self, remove_fields: List[str] = None) -> dict:
    copied = {}
    for key, val in self.__dict__.items():
      if remove_fields and key in remove_fields:
        continue

      if val:
        copied[key] = val

    return copied

  def to_group_and_artifact_only(self) -> 'MavenArtifact':
    return MavenArtifact(self.group_id, self.artifact_id)

  def to_file_name(self) -> str:
    return "%s-%s.%s" % (self.artifact_id, self.version, self.packaging)

  def to_path(
    self,
    include_version: bool = True,
    include_packaging: bool = True,
  ) -> str:
    parts = [self.group_id.replace(".", "/"), self.artifact_id]

    if include_version and self.version:
      parts.append(self.version)

      if include_packaging and self.packaging:
        parts.append(self.to_file_name())

    return "/".join(remove_nulls(parts))

  @staticmethod
  def from_artifact_string(artifact: str, strict: bool = True, star_to_null: bool = True):
    artifact_parts = artifact.split(":")

    parts_len = len(artifact_parts)
    if parts_len < 2 and strict:
      raise Exception("Artifact should be in the form of groupId:artifactId[:version[:packaging[:classifier[:scope]]]]")

    def get_part(idx: int) -> Optional[str]:
      if parts_len <= idx:
        return None

      part = artifact_parts[idx]

      if star_to_null and part == "*":
        return None

      return part

    return MavenArtifact(
      group_id=get_part(0),
      artifact_id=get_part(1),
      version=get_part(2),
      packaging=get_part(3),
      classifier=get_part(4),
      scope=get_part(5),
    )


def get_artifact_repository_path(artifact: Union[MavenArtifact, str]) -> Path:
  import os

  if isinstance(artifact, str):
    artifact = MavenArtifact.from_artifact_string(artifact)

  return Path(os.getenv("HOME")).joinpath(".m2/repository").joinpath(artifact.to_path(include_version=False, include_packaging=False))


def get_project_artifact_id(pom: Union[str, Path], use_mvn_expression: bool = False) -> str:
  result: str = None

  if not use_mvn_expression:
    result = xmlparser.parse_xml(pom).findtext('artifactId')

  if use_mvn_expression or not result:
    result = run_mvn_expression(pom, MVN_EXPR_PROJECT_ARTIFACT_ID)

  return result


def get_project_group_id(pom: Union[str, Path], use_mvn_expression: bool = False) -> str:
  result: str = None

  if not use_mvn_expression:
    parsed = xmlparser.parse_xml(pom)
    result = parsed.findtext('groupId')
    if not result:
      result = parsed.findtext('./parent/groupId')

  if use_mvn_expression or not result:
    result = run_mvn_expression(pom, MVN_EXPR_PROJECT_GROUP_ID)

  return result


def get_project_version(pom: Union[str, Path], use_mvn_expression: bool = False) -> str:
  result: str = None

  if not use_mvn_expression:
    parsed = xmlparser.parse_xml(pom)
    result = parsed.findtext('version')
    if not result:
      result = parsed.findtext('./parent/version')

  if use_mvn_expression or not result:
    result = run_mvn_expression(pom, MVN_EXPR_PROJECT_VERSION)

  return result


def ensure_artifact_version(artifact: MavenArtifact, rows: int = DEFAULT_SEARCH_ROWS, default_latest_version: bool = False) -> MavenArtifact:
  if not artifact.version:
    if default_latest_version:
      artifact.version = latest_artifact_version(artifact)
    else:
      artifact.version = select_artifact_version(artifact, rows=rows)

  return artifact


def ensure_artifact_packaging(artifact: MavenArtifact) -> MavenArtifact:
  if not artifact.packaging:
    artifact.packaging = select_artifact_packaging(artifact)

  return artifact


def latest_artifact_version(artifact: MavenArtifact) -> Optional[str]:
  search_result = call_latest_version_api(artifact)
  docs = search_result["response"]["docs"]
  if not docs:
    return None

  return docs[0]["latestVersion"]


# see https://central.sonatype.org/search/rest-api-guide/
def call_latest_version_api(artifact: MavenArtifact) -> dict:
  if not artifact.artifact_id:
    raise ValueError(f"artifact_id must be specified: {artifact}")

  query = f"g:{artifact.group_id} AND a:{artifact.artifact_id}"

  params = {
    "q": query,
    "rows": 1,
    "wt": "json",
  }
  return maybe_throw(requests.get("https://search.maven.org/solrsearch/select", params=params)).json()


# see https://central.sonatype.org/search/rest-api-guide/
def call_search_api(artifact: MavenArtifact, rows: int = DEFAULT_SEARCH_ROWS) -> dict:
  query = f"g:{artifact.group_id}"
  if artifact.artifact_id:
    query += f" AND a:{artifact.artifact_id}"

  params = {
    "q": query,
    "rows": rows,
    "wt": "json",
    "core": "gav",
  }
  return maybe_throw(requests.get("https://search.maven.org/solrsearch/select", params=params)).json()


def call_search_api_query(query: str, rows: int = DEFAULT_SEARCH_ROWS) -> dict:
  params = {
    "q": query,
    "rows": rows,
    "wt": "json",
    "core": "gav",
  }
  return maybe_throw(requests.get("https://search.maven.org/solrsearch/select", params=params)).json()


def call_remote_content_api(artifact: MavenArtifact) -> requests.Response:
  ensure_artifact_version(artifact)
  ensure_artifact_packaging(artifact)

  return maybe_throw(requests.get("https://search.maven.org/remotecontent?filepath=" + artifact.to_path()))


def select_artifact_version(artifact: MavenArtifact, rows: int = DEFAULT_SEARCH_ROWS) -> str:
  search_result = call_search_api(artifact, rows=rows)
  versions: List[str] = [doc.get("v") for doc in search_result["response"]["docs"]]
  return select_prompt(versions)


def select_artifact_packaging(artifact: MavenArtifact) -> str:
  return select_prompt(["jar", "pom"])


def parse_mvn_expression_output(mvn_out: str) -> str:
  return [line for line in mvn_out.splitlines() if '[INFO]' not in line][-1]


def run_mvn_expression(pom: Union[str, Path], expression: str) -> str:
  if isinstance(pom, str):
    pom = Path(pom)

  mvn_out = procs.run_and_parse_output(
    ['mvn', '-f', pom.absolute().as_posix(), 'help:evaluate', '-Dexpression={0}'.format(expression), '--quiet', '-DforceStdout'],
    check=True,
  )[1].strip()
  return mvn_out


def update_dep_version(pom: Union[str, Path], dep: str, version: str, additional_arguments: List[str] = None, check: bool = False) -> subprocess.CompletedProcess:
  if isinstance(pom, str):
    pom = Path(pom)

  if dep.count(':') <= 1:
    dep += ':*'

  mvn_command = [
    'mvn',
    '-f',
    pom.absolute().as_posix(),
    'versions:use-dep-version',
    '-DdepVersion=%s' % version,
    '-DgenerateBackupPoms=false',
    '-Dincludes=%s' % dep,
    '-DallowSnapshots=true',
  ]
  if additional_arguments:
    mvn_command.extend(additional_arguments)

  return procs.run(mvn_command, check=check)


def update_project_version(pom: Union[str, Path], version: str, additional_arguments: List[str] = None, check: bool = False) -> Tuple[int, str]:
  if isinstance(pom, str):
    pom = Path(pom)

  mvn_command = [
    'mvn',
    '-f',
    pom.absolute().as_posix(),
    'versions:set',
    '-DgenerateBackupPoms=false',
    '-DnewVersion=' + version,
  ]
  if additional_arguments:
    mvn_command.extend(additional_arguments)

  return procs.run_and_parse_output(mvn_command, check=check)


def _main():
  import sys

  result = globals()[sys.argv[1]](*sys.argv[2:])
  if result is not None:
    print(result)


if __name__ == "__main__":
  try:
    _main()
  except KeyboardInterrupt:
    exit(130)



#!/usr/bin/env python

from concurrent import futures
from typing import Callable, Iterable, List, TypeVar

T = TypeVar('T')
V = TypeVar('V')


def trap_pool_shutdown(pool: futures.Executor, wait: bool = False, cancel_futures: bool = True):
  import signal

  def shutdown_handler():
    pool.shutdown(wait=wait, cancel_futures=cancel_futures)

  signal.signal(signal.SIGINT, shutdown_handler)
  signal.signal(signal.SIGTERM, shutdown_handler)


def run_parallel(
  fn: Callable[[T], V],
  *iterables: Iterable[T],
  timeout: float = None,
  chunksize: int = 1,
  max_workers: int = None,
  use_io_optimized_pool: bool = True,
  pool: futures.Executor = None,
) -> List[V]:
  if pool is not None:
    return list(pool.map(fn, *iterables, timeout=timeout, chunksize=chunksize))

  if use_io_optimized_pool:
    from multiprocessing.pool import Pool

    with Pool(max_workers) as pool:
      return list(pool.map(fn, *iterables, chunksize=chunksize))

  with futures.ProcessPoolExecutor(max_workers=max_workers) as pool:
    return list(pool.map(fn, *iterables, timeout=timeout, chunksize=chunksize))



#!/usr/bin/env python
import re
import sqlite3
from pathlib import Path
from typing import Any, Dict, List, Union

import sqlalchemy
import sqlalchemy.engine.url

from ltpylib import configs, files, patterns, procs
from ltpylib.common_types import DataWithUnknownPropertiesAsAttributes

DEFAULT_PG_SERVICE_CONFIG_SECTION = "dwh"
PG_ENGINES: Dict[str, sqlalchemy.engine.Engine] = {}

SQL_CMD_REGEX_MAIN = r"((SELECT|WITH|EXPLAIN)[^;]*?^;)"
SQL_CMD_REGEX_PRIMARY = r"(?s)^-- ?use\n" + SQL_CMD_REGEX_MAIN
SQL_CMD_REGEX_SECONDARY = r"(?s)^" + SQL_CMD_REGEX_MAIN
SQL_CMD_REGEX_QUERY_ID_REPL_STR = "<query_id>"
SQL_CMD_REGEX_QUERY_ID = r"(?s)^-- ?" + SQL_CMD_REGEX_QUERY_ID_REPL_STR + r"\n([^;]+^;)"
SQL_CMD_REGEX_QUERY_ID_ALL = r"(?s)^-- ?([a-zA-Z0-9_-]+)\n([^;]+^;)"
SQL_CMD_REGEX_FLAGS = re.MULTILINE
SQL_CMD_REGEX_GROUP = 1


class PgServiceConfig(DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.dbname: str = values.pop("dbname", None)
    self.host: str = values.pop("host", None)
    self.password: str = values.pop("password", None)
    self.port: int = int(values.pop("port")) if "port" in values else None
    self.user: str = values.pop("user", None)

    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


def create_sqlite_connection(
  db_file: Union[Path, str],
  detect_types: int = sqlite3.PARSE_DECLTYPES,
  use_row_factory_as_dict: bool = True,
) -> sqlite3.Connection:
  db_conn = sqlite3.connect(
    db_file,
    detect_types=detect_types,
  )

  if use_row_factory_as_dict:
    db_conn.row_factory = sqlite_row_factory_as_dict

  return db_conn


def sqlite_row_factory_as_dict(cursor: sqlite3.Cursor, row) -> Dict[str, Any]:
  row_as_dict = {}
  for idx, col in enumerate(cursor.description):
    row_as_dict[col[0]] = row[idx]
  return row_as_dict


def parse_pg_service_config_file(section: str = None) -> PgServiceConfig:
  config_file = Path.home().joinpath(".pg_service.conf")
  if not config_file.is_file():
    raise ValueError(".pg_service.conf file does not exist at: %s" % config_file.as_posix())

  use_mock_default_section = section is None
  parsed = configs.read_properties(config_file, use_mock_default_section=use_mock_default_section)

  if use_mock_default_section:
    parsed_as_dict = {key: val for key, val in parsed.defaults()}
  else:
    parsed_as_dict = {key: val for key, val in parsed.items(section)}

  return PgServiceConfig(values=parsed_as_dict)


def pg_query(
  sql: str,
  *multi_params,
  config: PgServiceConfig = None,
  **params,
) -> sqlalchemy.engine.ResultProxy:
  if params is not None:
    params = convert_pg_params_to_correct_types(params)

  engine = get_or_create_pg_engine(config if config else parse_pg_service_config_file(DEFAULT_PG_SERVICE_CONFIG_SECTION))
  return engine.execute(sqlalchemy.sql.text(sql), *multi_params, **params)


def pg_query_to_dicts(
  sql: str,
  *multi_params,
  config: PgServiceConfig = None,
  **params,
) -> List[Dict[str, Any]]:
  return query_result_to_dicts(pg_query(sql, *multi_params, config=config, **params))


def query_result_to_dicts(result: sqlalchemy.engine.ResultProxy) -> List[Dict[str, Any]]:
  return [dict(row.items()) for row in result.fetchall()]


def convert_pg_params_to_correct_types(params: dict) -> dict:
  for key, val in params.items():
    if isinstance(val, list):
      params[key] = tuple(val)

  return params


def create_pg_engine(config: PgServiceConfig) -> sqlalchemy.engine.Engine:
  db_connect_url = sqlalchemy.engine.url.URL(
    drivername="postgresql+psycopg2",  # pg+psycopg2
    username=config.user,
    password=config.password,
    host=config.host,
    port=config.port,
    database=config.dbname,
  )
  return sqlalchemy.create_engine(db_connect_url)


def get_or_create_pg_engine(config: PgServiceConfig) -> sqlalchemy.engine.Engine:
  if str(config) not in PG_ENGINES:
    PG_ENGINES[str(config)] = create_pg_engine(config)

  return PG_ENGINES[str(config)]


def pull_sql_from_file(query_ids: Union[List[str], str], resolved_sql_file: Path, inject_env: bool = True) -> str:
  if isinstance(query_ids, str):
    query_ids = [query_ids]

  if query_ids:
    queries: List[str] = []
    for qid in query_ids:
      queries.append(match_query_regex_in_file(
        resolved_sql_file,
        [
          SQL_CMD_REGEX_QUERY_ID.replace(SQL_CMD_REGEX_QUERY_ID_REPL_STR, qid),
        ],
      ))

    query = "\n".join(queries)

  else:
    query = match_query_regex_in_file(
      resolved_sql_file,
      [
        SQL_CMD_REGEX_PRIMARY,
        SQL_CMD_REGEX_SECONDARY,
      ],
    )

  if inject_env:
    exit_code, updated_query = procs.run_and_parse_output(["envsubst"], input=query)
    if exit_code == 0:
      query = updated_query

  return query


def pull_query_ids_from_file(sql_file: Path) -> List[str]:
  return patterns.pull_matches_from_file(sql_file, SQL_CMD_REGEX_QUERY_ID_ALL, group=1, flags=SQL_CMD_REGEX_FLAGS)


def locate_sql_file(
  sql_file: str,
  base_dir: Path = None,
  default_file: Path = None,
) -> Path:
  if not sql_file:
    if not default_file:
      raise ValueError("You must either pass --default-file or a valid <sql_file> parameter to this script.")

    return default_file

  for maybe_file in [Path(sql_file), Path(sql_file + ".sql")]:
    if maybe_file.is_file():
      return maybe_file

  if not base_dir:
    raise ValueError("<sql_file> does not exist: " + sql_file)

  for maybe_file in [base_dir.joinpath(sql_file), base_dir.joinpath(sql_file + ".sql")]:
    if maybe_file.is_file():
      return maybe_file

  raise ValueError("<sql_file> does not exist: " + sql_file)


def match_query_regex_in_file(resolved_sql_file: Path, regexes_to_check: List[str]) -> str:
  file_contents = files.read_file(resolved_sql_file)

  for regex_to_check in regexes_to_check:
    for match in re.finditer(regex_to_check, file_contents, flags=SQL_CMD_REGEX_FLAGS):
      return match.group(SQL_CMD_REGEX_GROUP)

  raise Exception("Could not find a sql query match in file: " + resolved_sql_file.as_posix())



#!/usr/bin/env python
from decimal import Decimal, ROUND_HALF_DOWN
from typing import Union


def convert_decimal_precision(val: Union[Decimal, float, str], precision: int, rounding: str = ROUND_HALF_DOWN) -> Decimal:
  if precision == 0:
    exp = Decimal("1")
  else:
    exp = Decimal("." + ("0" * (precision - 1)) + "1")

  if not isinstance(val, Decimal):
    val = Decimal(val)

  return val.quantize(
    exp,
    rounding=rounding,
  )


def format_number(number: Union[Decimal, float, int], precision: int = 0, default_value: str = "") -> str:
  if number is None:
    return default_value

  format_str = "{:,.%sf}" % precision
  return format_str.format(number)



#!/usr/bin/env python
from typing import Union

from requests import Response


class NotFoundException(Exception):

  def __init__(self, response):
    try:
      self.data = response.json()
      if "errors" in self.data:
        msg = self.data["errors"][0]["message"]
      else:
        msg = str(self.data)
    except ValueError:
      msg = "Not found: " + response.url

    super(NotFoundException, self).__init__(msg)


class GenericException(Exception):

  def __init__(self, response):
    try:
      self.data = response.json()
      msg = "%d: %s" % (response.status_code, self.data)
    except ValueError:
      msg = "Unknown error: %d(%s)" % (response.status_code, response.reason)

    super(GenericException, self).__init__(msg)


class AuthenticationException(Exception):

  def __init__(self, response):
    try:
      msg = "%d: Invalid User / Password" % response.status_code
    except ValueError:
      msg = "Invalid Authentication"

    super(AuthenticationException, self).__init__(msg)


def add_json_headers(kw: dict = None) -> dict:
  if kw is None:
    kw = {}

  if "headers" not in kw:
    kw["headers"] = {}

  headers = {
    "Content-type": "application/json",
    "Accept": "application/json",
  }

  for header, value in headers.items():
    kw["headers"][header] = value

  return kw


def maybe_throw(response: Response) -> Response:
  if not response.ok:
    if response.status_code == 404:
      raise NotFoundException(response)
    elif response.status_code == 401:
      raise AuthenticationException(response)
    else:
      e = GenericException(response)
      try:
        e.data = response.json()
      except ValueError:
        e.content = response.content
      raise e

  return response


def parse_raw_response(response: Response) -> Union[dict, list, str]:
  maybe_throw(response)
  try:
    return response.json()
  except ValueError:
    return response.text



#!/usr/bin/env python
from typing import Any, List, Optional, Tuple

import jenkinsapi
import requests
from requests import Session

from ltpylib import output, requests_helper
from ltpylib.jenkins import create_recursive_tree_param
from ltpylib.jenkins_types import JenkinsBuild, JenkinsInstance


class JenkinsApi(object):

  def __init__(self, base_url: str, creds: Tuple[str, str]):
    if base_url.endswith("/"):
      self.base_url: str = base_url[:-1]
    else:
      self.base_url: str = base_url

    self.creds: Tuple[str, str] = creds
    self.session: Session = requests.Session()

    self.session.verify = True
    if creds is not None:
      self.session.auth = creds

    self.session.cookies = self.session.head(self.url("")).cookies
    self.session.headers.update({'Content-Type': 'application/json', 'Accept': 'application/json'})

    self._api: Optional[jenkinsapi.jenkins.Jenkins] = None

  @property
  def api(self) -> jenkinsapi.jenkins.Jenkins:
    if not self._api:
      self._api = jenkinsapi.jenkins.Jenkins(
        self.base_url,
        username=self.creds[0] if self.creds else None,
        password=self.creds[1] if self.creds else None,
        lazy=True,
      )

    return self._api

  def url(self, resource_path: str):
    if resource_path.startswith("https://") or resource_path.startswith("http://"):
      return resource_path

    if not resource_path.startswith("/"):
      resource_path = "/" + resource_path

    return self.base_url + resource_path

  def all_builds(
    self,
    job: str,
    tree: str = "building,description,displayName,duration,estimatedDuration,executor,id,number,queueId,result,timestamp,url",
  ) -> List[JenkinsBuild]:
    response = self.session.get(self.url("job/%s/api/json?tree=%s" % (
      job,
      "allBuilds[%s]" % tree,
    )))
    return [JenkinsBuild(values=v) for v in requests_helper.parse_raw_response(response).get("allBuilds")]

  def build(
    self,
    job: str,
    build: int,
  ) -> JenkinsBuild:
    response = self.session.get(self.url("job/%s/%s/api/json?tree=*" % (job, str(build))))
    return JenkinsBuild(requests_helper.parse_raw_response(response))

  def build_from_full_url(
    self,
    full_url: str,
  ) -> JenkinsBuild:
    response = self.session.get(self.url("%s/api/json?tree=*" % full_url))
    return JenkinsBuild(requests_helper.parse_raw_response(response))

  def instance_info(self, tree: str = "*") -> JenkinsInstance:
    response = self.session.get(self.url("api/json?tree=%s" % tree))
    return JenkinsInstance(requests_helper.parse_raw_response(response))

  def instance_info_all_jobs(
    self,
    fields: List[str] = ("color", "name", "url"),
    depth: int = 10,
  ) -> JenkinsInstance:
    return self.instance_info(tree=create_recursive_tree_param("jobs", fields, depth))

  def job(self, job_url_or_path: str) -> jenkinsapi.job.Job:
    job_url = job_url_or_path if job_url_or_path.startswith(self.base_url) else f"{self.base_url}/job/{job_url_or_path}"
    job_name = jenkinsapi.job.Job.get_full_name_from_url_and_baseurl(job_url, self.base_url)
    return self.api.get_job_by_url(job_url, job_name)


def jenkinsapi_dumper(val: Any) -> Any:
  return val._data


def jenkinsapi_dumper_use_if(val: Any) -> bool:
  return isinstance(val, jenkinsapi.jenkinsbase.JenkinsBase)


def create_default_jenkins_api(
  url: str = None,
  user: str = None,
  token: str = None,
) -> JenkinsApi:
  if not url:
    import os

    url = os.getenv("JENKINS_BASE_URL")

  return JenkinsApi(url, create_jenkins_auth(user=user, token=token))


def create_jenkins_auth(user: str = None, token: str = None) -> Tuple[str, str]:
  if not user:
    import os

    user = os.getenv("JENKINS_USER_ID")

  if not token:
    import os

    token = os.getenv("JENKINS_API_TOKEN")

  return user, token


output.add_custom_json_dumper("jenkinsapi", jenkinsapi_dumper, use_if=jenkinsapi_dumper_use_if)



#!/usr/bin/env python
# pylint: disable=C0111
import re
from pathlib import Path
from typing import Callable, List, Match, Sequence, Union


def join_by_regex_or(parts: Sequence[str]) -> str:
  return "(" + "|".join(parts) + ")"


def replace_matches(
  content: str,
  search_string: Union[str, re.Pattern],
  replacement: Union[str, Callable[[Match], str]],
  quote_replacement: Union[bool, str] = False,
  wrap_replacement_in_function: bool = False,
  flags: Union[int, re.RegexFlag] = 0,
) -> str:
  if isinstance(quote_replacement, str):
    quote_replacement = quote_replacement.lower() in ['true', '1', 't', 'y', 'yes']

  if quote_replacement and isinstance(replacement, str):
    replacement = re.escape(replacement)
  elif wrap_replacement_in_function and isinstance(replacement, str):
    replacement_content = replacement

    def replacement_function(match: Match) -> str:
      return replacement_content

    replacement = replacement_function

  return re.sub(search_string, replacement, content, flags=flags)


def pull_matches_from_file(
  file: Union[str, Path],
  search_string: Union[str, re.Pattern],
  group: int = 0,
  flags: Union[int, re.RegexFlag] = 0,
) -> List[str]:
  from ltpylib import files

  content = files.read_file(file)

  matches: List[str] = []
  for match in re.finditer(search_string, content, flags=flags):
    matches.append(match.group(group))

  return matches


def regex_match(val: str, patterns: List[Union[str, re.Pattern]]) -> bool:
  for pattern in patterns:
    if re.fullmatch(pattern, val):
      return True

  return False


def _main():
  import sys

  result = globals()[sys.argv[1]](*sys.argv[2:])
  if result is not None:
    print(result)


if __name__ == "__main__":
  try:
    _main()
  except KeyboardInterrupt:
    exit(130)



#!/usr/bin/env python
# pylint: disable=C0111
import inspect
from typing import ClassVar, List, Tuple, Type

SIMPLE_TYPES: Tuple[Type] = (str, bool, int, float)


def create_typing_description(val) -> str:
  if isinstance(val, list):
    if len(val) == 0:
      return "List"
    else:
      return "List[%s]" % create_typing_description(val[0])
  else:
    return type_name(val)


def type_name(val) -> str:
  return type(val).__name__


def get_functions_of_class(
  inspect_class: ClassVar,
  include_private: bool = False,
  include_signature: bool = False,
) -> List[str]:
  result: List[str] = []
  functions = inspect.getmembers(inspect_class, predicate=inspect.isfunction)
  for func in functions:
    if not include_private and func[0].startswith("_"):
      continue

    func_info: str = func[0]

    if include_signature:
      func_sig: str = str(inspect.signature(func[1])).replace("(self, ", "(", 1).replace("(self)", "()", 1)
      func_info += func_sig

    result.append(func_info)

  return result



#!/usr/bin/env python
import argparse
from enum import auto
from typing import Dict, List

from ltpylib import enums, opts_actions
from ltpylib.common_types import DataWithUnknownProperties, TypeWithDictRepr


class DisplayIdAndId(object):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.displayId: str = values.pop("displayId", None)
    self.id: str = values.pop("id", None)


class PaginatedValues(object):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.isLastPage: int = values.pop("isLastPage", None)
    self.limit: int = values.pop("limit", None)
    self.size: int = values.pop("size", None)
    self.start: int = values.pop("start", None)


class PullRequestParticipantStatus(enums.EnumAutoName):
  APPROVED = auto()
  NEEDS_WORK = auto()
  UNAPPROVED = auto()

  @staticmethod
  def from_string(val: str, allow_unknown: bool = False):
    try:
      return PullRequestParticipantStatus[val.upper()] if val else None
    except KeyError as e:
      if allow_unknown:
        return None

      raise e


class PullRequestRole(enums.EnumAutoName):
  AUTHOR = auto()
  PARTICIPANT = auto()
  REVIEWER = auto()

  @staticmethod
  def from_string(val: str, allow_unknown: bool = False):
    try:
      return PullRequestRole[val.upper()] if val else None
    except KeyError as e:
      if allow_unknown:
        return None

      raise e


class PullRequestState(enums.EnumAutoName):
  DECLINED = auto()
  MERGED = auto()
  OPEN = auto()
  UNKNOWN = auto()

  @staticmethod
  def from_string(state: str, allow_unknown: bool = True, unknown_as_none: bool = False):
    try:
      return PullRequestState[state.upper()] if state else None
    except KeyError as e:
      if allow_unknown:
        return None if unknown_as_none else PullRequestState.UNKNOWN

      raise e


class ApplicationProperties(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.buildDate: str = values.pop("buildDate", None)
    self.buildNumber: str = values.pop("buildNumber", None)
    self.displayName: str = values.pop("displayName", None)
    self.version: str = values.pop("version", None)

    DataWithUnknownProperties.__init__(self, values)


class Branch(DataWithUnknownProperties, DisplayIdAndId):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    DisplayIdAndId.__init__(self, values)

    self.isDefault: bool = values.pop("isDefault", None)
    self.latestChangeset: str = values.pop("latestChangeset", None)
    self.latestCommit: str = values.pop("latestCommit", None)
    self.type: str = values.pop("type", None)
    self.metadata: Dict[str, dict] = values.pop("metadata", None)

    DataWithUnknownProperties.__init__(self, values)


class Branches(DataWithUnknownProperties, PaginatedValues):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    PaginatedValues.__init__(self, values)

    self.values: List[Branch] = list(map(Branch, values.pop("values", []))) if "values" in values else None

    DataWithUnknownProperties.__init__(self, values)


class Build(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.dateAdded: int = values.pop("dateAdded", None)
    self.description: str = values.pop("description", None)
    self.key: str = values.pop("key", None)
    self.name: str = values.pop("name", None)
    self.state: str = values.pop("state", None)
    self.url: str = values.pop("url", None)

    DataWithUnknownProperties.__init__(self, values)


class Builds(DataWithUnknownProperties, PaginatedValues):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    PaginatedValues.__init__(self, values)

    self.values: List[Build] = list(map(Build, values.pop("values", []))) if "values" in values else None

    DataWithUnknownProperties.__init__(self, values)


class Comment(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.id: int = values.pop("id", None)
    self.text: str = values.pop("text", None)
    self.version: int = values.pop("version", None)

    DataWithUnknownProperties.__init__(self, values)


class Commit(DataWithUnknownProperties, DisplayIdAndId):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    DisplayIdAndId.__init__(self, values)

    DataWithUnknownProperties.__init__(self, values)


class Link(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.href: str = values.pop("href", None)
    self.name: str = values.pop("name", None)

    DataWithUnknownProperties.__init__(self, values)


class Links(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.clone: List[Link] = list(map(Link, values.pop("clone", []))) if "clone" in values else None
    self.self: List[Link] = list(map(Link, values.pop("self", []))) if "self" in values else None

    DataWithUnknownProperties.__init__(self, values)


class Project(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.description: str = values.pop("description", None)
    self.id: int = values.pop("id", None)
    self.key: str = values.pop("key", None)
    self.links: Links = Links(values=values.pop("links")) if "links" in values else None
    self.name: str = values.pop("name", None)
    self.owner: User = User(values=values.pop("owner")) if "owner" in values else None
    self.public: bool = values.pop("public", None)
    self.type: str = values.pop("type", None)

    DataWithUnknownProperties.__init__(self, values)


class PullRequestActivities(DataWithUnknownProperties, PaginatedValues):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    PaginatedValues.__init__(self, values)

    self.values: List[PullRequestActivity] = list(map(PullRequestActivity, values.pop("values", []))) if "values" in values else None

    DataWithUnknownProperties.__init__(self, values)


class PullRequestActivity(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.action: str = values.pop("action", None)
    self.commentAction: str = values.pop("commentAction", None)
    self.createdDate: int = values.pop("createdDate", None)
    self.id: int = values.pop("id", None)
    self.user: User = User(values=values.pop("user")) if "user" in values else None

    DataWithUnknownProperties.__init__(self, values)


class PullRequestMergeability(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.canMerge: bool = values.pop("canMerge", None)
    self.conflicted: bool = values.pop("conflicted", None)
    self.outcome: str = values.pop("outcome", None)
    self.vetoes: List[PullRequestVeto] = list(map(PullRequestVeto, values.pop("vetoes", []))) if "vetoes" in values else None

    DataWithUnknownProperties.__init__(self, values)


class PullRequestParticipant(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.approved: bool = values.pop("approved", None)
    self.lastReviewedCommit: str = values.pop("lastReviewedCommit", None)
    self.role: str = values.pop("role", None)
    self.status: str = values.pop("status", None)
    self.user: User = User(values=values.pop("user")) if "user" in values else None

    DataWithUnknownProperties.__init__(self, values)


class PullRequestProperties(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.commentCount: int = values.pop("commentCount", None)
    self.mergeCommit: Commit = Commit(values=values.pop("mergeCommit")) if "mergeCommit" in values else None
    self.openTaskCount: int = values.pop("openTaskCount", None)
    self.resolvedTaskCount: int = values.pop("resolvedTaskCount", None)

    DataWithUnknownProperties.__init__(self, values)


class PullRequestRef(DataWithUnknownProperties, DisplayIdAndId):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    DisplayIdAndId.__init__(self, values)

    self.latestCommit: str = values.pop("latestCommit", None)
    self.repository: Repository = Repository(values=values.pop("repository")) if "repository" in values else None

    DataWithUnknownProperties.__init__(self, values)


class PullRequestStatus(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.author: PullRequestParticipant = PullRequestParticipant(values=values.pop("author")) if "author" in values else None
    self.closed: bool = values.pop("closed", None)
    self.createdDate: int = values.pop("createdDate", None)
    self.description: str = values.pop("description", None)
    self.fromRef: PullRequestRef = PullRequestRef(values=values.pop("fromRef")) if "fromRef" in values else None
    self.id: int = values.pop("id", None)
    self.links: Links = Links(values=values.pop("links")) if "links" in values else None
    self.locked: bool = values.pop("locked", None)
    self.open: bool = values.pop("open", None)
    self.outstandingTaskCount: int = values.pop("outstandingTaskCount", None)
    self.participants: List[PullRequestParticipant] = list(map(PullRequestParticipant, values.pop("participants", []))) if "participants" in values else None
    self.properties: PullRequestProperties = PullRequestProperties(values=values.pop("properties")) if "properties" in values else None
    self.reviewers: List[PullRequestParticipant] = list(map(PullRequestParticipant, values.pop("reviewers", []))) if "reviewers" in values else None
    self.state: PullRequestState = PullRequestState.from_string(values.pop("state")) if "state" in values else None
    self.title: str = values.pop("title", None)
    self.toRef: PullRequestRef = PullRequestRef(values=values.pop("toRef")) if "toRef" in values else None
    self.updatedDate: int = values.pop("updatedDate", None)
    self.version: int = values.pop("version", None)

    self.url: str = self.links.self[0].href if (self.links and self.links.self) else None

    self.builds: Builds = None
    self.mergeInfo: PullRequestMergeability = None

    DataWithUnknownProperties.__init__(self, values)


class PullRequestStatuses(DataWithUnknownProperties, PaginatedValues):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    PaginatedValues.__init__(self, values)

    self.nextPageStart: int = values.pop("nextPageStart", None)
    self.values: List[PullRequestStatus] = list(map(PullRequestStatus, values.pop("values", []))) if "values" in values else None

    DataWithUnknownProperties.__init__(self, values)


class PullRequestMergeStatus(PullRequestStatus):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.sourceBranchDeleted: bool = None

    PullRequestStatus.__init__(self, values)


class PullRequestVeto(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.detailedMessage: str = values.pop("detailedMessage", None)
    self.summaryMessage: str = values.pop("summaryMessage", None)

    DataWithUnknownProperties.__init__(self, values)


class Repository(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.forkable: bool = values.pop("forkable", None)
    self.id: int = values.pop("id", None)
    self.links: Links = Links(values=values.pop("links")) if "links" in values else None
    self.name: str = values.pop("name", None)
    self.origin: Repository = Repository(values=values.pop("origin")) if "origin" in values else None
    self.project: Project = Project(values=values.pop("project")) if "project" in values else None
    self.public: bool = values.pop("public", None)
    self.scmId: str = values.pop("scmId", None)
    self.slug: str = values.pop("slug", None)
    self.state: str = values.pop("state", None)
    self.statusMessage: str = values.pop("statusMessage", None)

    DataWithUnknownProperties.__init__(self, values)


class SearchGroupResult(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.category: str = values.pop("category", None)
    self.count: int = values.pop("count", None)
    self.isLastPage: int = values.pop("isLastPage", None)
    self.nextStart: int = values.pop("nextStart", None)
    self.start: int = values.pop("start", None)
    self.values: List[SearchResult] = list(map(SearchResult, values.pop("values", []))) if "values" in values else None

    DataWithUnknownProperties.__init__(self, values)


class SearchResult(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.file: str = values.pop("file", None)
    self.hitContexts: List[List[SearchResultHit]] = list(map(lambda vals: list(map(SearchResultHit, vals)), values.pop("hitContexts", []))) if "hitContexts" in values else None
    self.hitCount: int = values.pop("hitCount", None)
    self.pathMatches: List[SearchResultPathMatch] = list(map(SearchResultPathMatch, values.pop("pathMatches", []))) if "pathMatches" in values else None
    self.repository: Repository = Repository(values=values.pop("repository")) if "repository" in values else None

    DataWithUnknownProperties.__init__(self, values)


class SearchResultHit(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.line: int = values.pop("line", None)
    self.text: str = values.pop("text", None)

    DataWithUnknownProperties.__init__(self, values)


class SearchResultPathMatch(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.match: bool = values.pop("match", None)
    self.text: str = values.pop("text", None)

    DataWithUnknownProperties.__init__(self, values)


class SearchResultQuery(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.substituted: bool = values.pop("substituted", None)

    DataWithUnknownProperties.__init__(self, values)


class SearchResultScope(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.type: str = values.pop("type", None)

    DataWithUnknownProperties.__init__(self, values)


class SearchResults(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.code: SearchGroupResult = SearchGroupResult(values=values.pop("code")) if "code" in values else None
    self.query: SearchResultQuery = SearchResultQuery(values=values.pop("query")) if "query" in values else None
    self.scope: SearchResultScope = SearchResultScope(values=values.pop("scope")) if "scope" in values else None

    DataWithUnknownProperties.__init__(self, values)


class Tag(DataWithUnknownProperties, DisplayIdAndId):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    DisplayIdAndId.__init__(self, values)

    self.hash: str = values.pop("hash", None)
    self.latestChangeset: str = values.pop("latestChangeset", None)
    self.latestCommit: str = values.pop("latestCommit", None)

    DataWithUnknownProperties.__init__(self, values)


class User(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.active: bool = values.pop("active", None)
    self.deletable: bool = values.pop("deletable", None)
    self.displayName: str = values.pop("displayName", None)
    self.emailAddress: str = values.pop("emailAddress", None)
    self.id: int = values.pop("id", None)
    self.lastAuthenticationTimestamp: int = values.pop("lastAuthenticationTimestamp", None)
    self.links: Links = Links(values=values.pop("links")) if "links" in values else None
    self.mutableDetails: bool = values.pop("mutableDetails", None)
    self.mutableGroups: bool = values.pop("mutableGroups", None)
    self.name: str = values.pop("name", None)
    self.slug: str = values.pop("slug", None)
    self.type: str = values.pop("type", None)

    DataWithUnknownProperties.__init__(self, values)


class ScriptArgPullRequest(TypeWithDictRepr):
  url_regex: str = r"https://([^/]+)/(projects|users)/([^/]+)/repos/([^/]+)/pull-requests/([^/]+)"

  def __init__(
    self,
    url: str = None,
    project: str = None,
    repo: str = None,
    pr_id: int = None,
    stash_url: str = None,
  ):
    self.url: str = url
    self.project: str = project
    self.repo: str = repo
    self.pr_id: int = pr_id

    self.project_type: str = ""

    if self.url:
      import re

      matches: List[List[str]] = re.findall(ScriptArgPullRequest.url_regex, self.url)
      if not matches:
        raise ValueError("Could not extract arguments from URL: %s" % self.url)

      matched = matches[0]
      self.pr_id: int = int(matched[4])
      self.project: str = matched[2]
      self.project_type: str = matched[1]
      self.repo: str = matched[3]

      if self.project_type == "users" and not self.project.startswith("~"):
        self.project = "~" + self.project

    else:
      if self.pr_id and self.project and self.repo:

        if self.project.startswith("~"):
          self.project_type = "users"
        else:
          self.project_type = "projects"

        if stash_url.endswith("/"):
          stash_url = stash_url[:-1]

        self.url = ("%s/projects/%s/repos/%s/pull-requests/%s/overview" % (stash_url, self.project, self.repo, self.pr_id))
      else:
        raise ValueError("Either the URL (via --url) or project/repo/id (via --project/--repo/--id) arguments must be specified.")


class ScriptArgsPullRequestSelections(object):

  def __init__(self, args: argparse.Namespace):
    self.participant_status: List[PullRequestParticipantStatus] = args.participant_status
    self.role: PullRequestRole = args.role
    self.state: PullRequestState = args.state
    self.state_all: bool = args.state_all

    if self.state_all:
      self.state = None

  @staticmethod
  def add_arguments_to_parser(arg_parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    arg_parser.add_argument("--participant-status", type=PullRequestParticipantStatus, choices=list(PullRequestParticipantStatus), nargs="*")
    arg_parser.add_argument("--role", type=PullRequestRole, choices=list(PullRequestRole))
    arg_parser.add_argument("--state", type=PullRequestState, choices=list(PullRequestState), default=PullRequestState.OPEN)
    arg_parser.add_argument("--state-all", action=opts_actions.STORE_TRUE)
    return arg_parser



#!/usr/bin/env python
import os
from typing import Tuple

from ltpylib.jira_api import JiraApi

JIRA_URL = os.getenv("JIRA_URL")
JIRA_TOKEN_ENV_VAR = "JIRA_API_TOKEN"
JIRA_TOKEN_USER_ENV_VAR = "JIRA_API_USER"


def create_jira_api(url: str = JIRA_URL, auth: Tuple[str, str] = None) -> JiraApi:
  return JiraApi(
    url=url,
    basic_auth=(auth if auth is not None else get_atlassian_token_auth()),
  )


def get_atlassian_token_auth(env_var_name: str = JIRA_TOKEN_ENV_VAR) -> Tuple[str, str]:
  return os.getenv(JIRA_TOKEN_USER_ENV_VAR), os.getenv(env_var_name)



#!/usr/bin/env python

import glob
import io
import os
from collections import OrderedDict
from configparser import ConfigParser, NoOptionError, NoSectionError
from pathlib import Path
from typing import Dict, List, Union

HOSTS_CONFIG: Union[ConfigParser, None] = None


def get_host(key: str, use_key_as_default: bool = True) -> str:
  hosts = get_or_create_hosts_config()

  if use_key_as_default:
    try:
      return hosts.get('DEFAULT', key)
    except NoOptionError:
      return key
    except NoSectionError:
      return key
  else:
    return hosts.get('DEFAULT', key)


def get_or_create_hosts_config() -> ConfigParser:
  global HOSTS_CONFIG
  if HOSTS_CONFIG is None:
    hosts_dir = os.path.abspath('%s/config/hosts' % os.getenv('DOTFILES', '%s/.dotfiles' % os.getenv('HOME')))
    HOSTS_CONFIG = ConfigParser()
    host_files = glob.glob(hosts_dir + '/*.properties')
    host_files.extend(glob.glob(hosts_dir + '/*/*.properties'))
    for hosts_file in host_files:
      HOSTS_CONFIG = read_properties(hosts_file, config=HOSTS_CONFIG)

  return HOSTS_CONFIG


def sort_config_parser_keys(config: ConfigParser) -> ConfigParser:
  if config._defaults:
    config._defaults = OrderedDict(sorted(config._defaults.items(), key=lambda t: t[0]))

  for section in config._sections:
    config._sections[section] = OrderedDict(sorted(config._sections[section].items(), key=lambda t: t[0]))

  config._sections = OrderedDict(sorted(config._sections.items(), key=lambda t: t[0]))
  return config


def config_parser_to_string(config: ConfigParser, sort_keys: bool = False):
  if sort_keys:
    config = sort_config_parser_keys(config)

  sio = io.StringIO()
  config.write(sio)
  return sio.getvalue()


def read_properties(
  file: Union[str, Path],
  use_mock_default_section: bool = True,
  config: ConfigParser = None,
  allow_missing_file: bool = False,
) -> ConfigParser:
  if isinstance(file, str):
    file = Path(file)

  if config is None:
    config = ConfigParser(allow_no_value=True)
    config.optionxform = str

  if not file.is_file():
    if not allow_missing_file:
      raise ValueError("File does not exist: %s" % file)
    return config

  if use_mock_default_section:
    with open(file.as_posix(), 'r') as configfile:
      config.read_string('[DEFAULT]\n' + configfile.read())
  else:
    config.read(file.as_posix())

  return config


def write_properties(config: ConfigParser, file: Union[str, Path], sort_keys: bool = False):
  if sort_keys:
    config = sort_config_parser_keys(config)

  if isinstance(file, str):
    file = Path(file)

  with open(file.as_posix(), 'w') as configfile:
    config.write(configfile)


def read_bash_env_vars_from_file(file: Union[str, Path], pass_env_vars: List[str] = None) -> Dict[str, str]:
  from ltpylib import procs

  env_vars: Dict[str, str] = {}

  if isinstance(file, str):
    file = Path(file)

  command = ["env", "-i"]
  include_env_vars = ["HOME"]
  if pass_env_vars:
    include_env_vars.extend(pass_env_vars)

  for var_name in include_env_vars:
    command.append("%s=%s" % (var_name, os.getenv(var_name, "")))

  command.extend(["bash", "-c", "source '%s' && env" % file.as_posix()])
  output = procs.run_and_parse_output(command, check=True)[1]

  for line in output.splitlines():
    (key, _, value) = line.partition("=")
    env_vars[key] = value

  for var_name in (include_env_vars + ["_", "SHLVL", "PWD"]):
    if var_name in env_vars:
      env_vars.pop(var_name)

  return env_vars


def _main():
  import sys

  result = globals()[sys.argv[1]](*sys.argv[2:])
  if result is not None:
    print(result)


if __name__ == "__main__":
  try:
    _main()
  except KeyboardInterrupt:
    exit(130)



#!/usr/bin/env python
import re
from typing import Sequence, Tuple

JENKINS_JOB_URL_REGEX: str = r'^(https?):\/\/([^\/:]+)(:[0-9]+)?(?:\/)(?:job|blue\/organizations\/jenkins)\/(([^\/]+)(?:\/(?:job|detail)\/([^\/]+))?)\/([0-9]+)\/' \
                             r'(?:pipeline\/?|display\/redirect)?$'


def create_recursive_tree_param(container: str, fields: Sequence[str], depth: int) -> str:
  tree_param = f"{container}[{','.join(fields)}"
  start_tree = tree_param
  end_tree = "]"
  for idx in range(depth):
    start_tree += "," + tree_param
    end_tree += "]"

  return start_tree + end_tree


def parse_job_url(url: str) -> Tuple[str, int]:
  match = re.match(JENKINS_JOB_URL_REGEX, url)
  job_name = match.group(5)
  if match.group(6) and match.group(5) != match.group(6):
    job_name = job_name + "/job/" + match.group(6)

  job_num = int(match.group(7))

  return job_name, job_num



#!/usr/bin/env python


class TypeWithDictRepr(object):

  def __repr__(self) -> str:
    return str(self.__dict__)

  def as_dict(self) -> dict:
    return dict(self.__dict__)


class DataWithUnknownProperties(TypeWithDictRepr):

  def __init__(self, values: dict = None, skip_field_if_no_unknown: bool = False):
    if not skip_field_if_no_unknown or values:
      self.unknown_properties: dict = values if values else None


class DataWithUnknownPropertiesAsAttributes(DataWithUnknownProperties):

  def __init__(self, values: dict = None):
    if values:
      self._has_unknown_properties: bool = True
      for item in values.items():
        setattr(self, str(item[0]), item[1])

      values.clear()

    DataWithUnknownProperties.__init__(self, values=None, skip_field_if_no_unknown=True)

  @property
  def has_unknown_properties(self) -> bool:
    return getattr(self, "_has_unknown_properties", False)



#!/usr/bin/env python
import unittest

from ltpylib import checks


class TestChecks(unittest.TestCase):

  def test_is_empty(self):
    assert checks.is_empty(None)
    assert checks.is_empty([])
    assert checks.is_empty({})

  def test_is_not_empty(self):
    assert checks.is_not_empty("abc")
    assert checks.is_not_empty([1])
    assert checks.is_not_empty(["abc"])
    assert checks.is_not_empty([{}])
    assert checks.is_not_empty([[]])
    assert checks.is_not_empty({"field": "value"})


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
import time
from typing import List

from ltpylib.common_types import DataWithUnknownPropertiesAsAttributes


class JenkinsBuild(DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self._class: str = values.pop("_class", None)
    self.building: bool = values.pop("building", None)
    self.description: str = values.pop("description", None)
    self.displayName: str = values.pop("displayName", None)
    self.duration: int = values.pop("duration", None)
    self.estimatedDuration: int = values.pop("estimatedDuration", None)
    self.executor: str = values.pop("executor", None)
    self.id: str = values.pop("id", None)
    self.number: int = values.pop("number", None)
    self.queueId: int = values.pop("queueId", None)
    self.result: str = values.pop("result", None)
    self.timestamp: int = values.pop("timestamp", None)
    self.url: str = values.pop("url", None)

    self.timeRunning: int = None
    if self.timestamp is not None:
      self.timeRunning = (int(time.time() * 1000) - self.timestamp)

    self.estimatedTimeRemaining: int = None
    if self.duration is not None and self.duration > 0 and self.estimatedDuration is not None and self.estimatedDuration > 0:
      self.estimatedTimeRemaining: int = self.estimatedDuration - self.duration
    elif self.estimatedDuration is not None and self.estimatedDuration > 0 and self.timeRunning is not None:
      self.estimatedTimeRemaining: int = self.estimatedDuration - self.timeRunning

    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class JenkinsInstance(DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self._class: str = values.pop("_class", None)
    self.jobs: List[JenkinsJob] = list(map(JenkinsJob, values.pop("jobs", []))) if "jobs" in values else None

    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class JenkinsJob(DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self._class: str = values.pop("_class", None)
    self.jobs: List[JenkinsJob] = list(map(JenkinsJob, values.pop("jobs", []))) if "jobs" in values else None
    self.name: str = values.pop("name", None)
    self.url: str = values.pop("url", None)

    DataWithUnknownPropertiesAsAttributes.__init__(self, values)


class JenkinsJobStats(DataWithUnknownPropertiesAsAttributes):

  def __init__(self, values: dict = None):
    values = values if values is not None else {}

    self.avg_duration_millis: float = values.pop("avg_duration_millis", None)
    self.name: str = values.pop("name", None)
    self.total: int = values.pop("total", None)

    DataWithUnknownPropertiesAsAttributes.__init__(self, values)



#!/usr/bin/env python
import dataclasses
import re
import textwrap
from pathlib import Path
from typing import Any, Dict, List, Union

from ltpylib import files
from ltpylib.numbers_util import convert_decimal_precision
from ltpylib.output import dicts_to_csv, prettify_sql

SQLITE_QUOTE_COL_REGEX = re.compile(r"[^a-zA-Z0-9_]")


@dataclasses.dataclass
class SQLiteColumn:
  name: str
  value_type: str
  has_nulls: bool = False

  @staticmethod
  def col_sort(col: 'SQLiteColumn') -> list:
    return [col.name]

  def create_col_name(self) -> str:
    name = self.name
    if SQLITE_QUOTE_COL_REGEX.match(name):
      name = '"' + name + '"'

    return name

  def create_update_blanks_to_null(self, table_name: str) -> str:
    col_name = self.create_col_name()
    return f"""
UPDATE {table_name}
SET {col_name} = NULL
WHERE {col_name} = '';
    """.strip()

  def to_create_column(self) -> str:
    col_def = f"{self.create_col_name()} {self.value_type}"
    if not self.has_nulls:
      col_def = col_def + " NOT NULL"

    return col_def

  def to_primary_key(self) -> str:
    return f"PRIMARY KEY ({self.create_col_name()})"


@dataclasses.dataclass
class SQLiteCreateTable:
  columns: List[SQLiteColumn]
  sql_cmd: str
  rows_as_csv: str = None
  csv_file: Path = None


def add_sqlite_columns_from_dicts(
  datas: List[Dict[str, Any]],
  sqlite_cols: List[SQLiteColumn],
  ignore_cols: List[re.Pattern] = None,
  only_cols: List[str] = None,
) -> List[SQLiteColumn]:
  cols_by_name: Dict[str, SQLiteColumn] = {col.name: col for col in sqlite_cols}
  existing = [col.name for col in sqlite_cols]
  has_nulls = set()
  add_cols: List[SQLiteColumn] = []
  attr_cols: List[SQLiteColumn] = []

  for idx, data in enumerate(datas):
    missing_existing_cols = list(cols_by_name.keys())

    for key, value in data.items():
      if key in missing_existing_cols:
        missing_existing_cols.remove(key)
      elif idx > 0:
        has_nulls.add(key)

      if ignore_cols and any([regex.fullmatch(key) is not None for regex in ignore_cols]):
        continue
      elif only_cols and key not in only_cols:
        continue

      if value is None:
        has_nulls.add(key)

        if key in cols_by_name:
          cols_by_name.get(key).has_nulls = True

        continue
      elif key in existing:
        continue

      existing.append(key)
      value_type = "TEXT"
      if isinstance(value, int):
        value_type = "INTEGER"
      elif isinstance(value, float):
        value_type = "REAL"

      col = SQLiteColumn(name=key, value_type=value_type, has_nulls=(key in has_nulls))
      cols_by_name[col.name] = col

      if key.startswith("attr_"):
        attr_cols.append(col)
      else:
        add_cols.append(col)

    if missing_existing_cols:
      has_nulls.update(missing_existing_cols)
      for col_name in missing_existing_cols:
        cols_by_name.get(col_name).has_nulls = True

  sqlite_cols.extend(sorted(add_cols, key=SQLiteColumn.col_sort))
  sqlite_cols.extend(sorted(attr_cols, key=SQLiteColumn.col_sort))

  return sqlite_cols


def sqlite_create_table_from_dicts(
  table_name: str,
  rows: List[dict],
  existing_cols: List[SQLiteColumn] = None,
  primary_key: SQLiteColumn = None,
  ignore_cols: List[re.Pattern] = None,
  only_cols: List[str] = None,
  additional_table_config: List[str] = None,
  load_file: Union[str, Path] = None,
  create_csv_from_rows: bool = True,
  create_csv_convert_booleans: bool = True,
) -> SQLiteCreateTable:
  sqlite_cols: List[SQLiteColumn] = existing_cols.copy() if existing_cols else []

  add_sqlite_columns_from_dicts(
    rows,
    sqlite_cols,
    ignore_cols=ignore_cols,
    only_cols=only_cols,
  )

  sql_cmd_parts = []

  cols_sql_statements = [col.to_create_column() for col in sqlite_cols]

  if primary_key:
    cols_sql_statements.append(primary_key.to_primary_key())

  if additional_table_config:
    cols_sql_statements.extend(additional_table_config)

  cols_sql = ",\n  ".join(cols_sql_statements)

  drop_create_sql = prettify_sql(textwrap.dedent(f"""\
    DROP TABLE IF EXISTS {table_name}
    ;

    CREATE TABLE {table_name} (
      {cols_sql}
    )
    ;
  """.rstrip()))

  sql_cmd_parts.append(drop_create_sql)

  if load_file:
    if isinstance(load_file, Path):
      load_file = load_file.as_posix()

    sqlite_cmds = textwrap.dedent(
      f"""\
      -- @formatter:off
      .mode csv
      .separator |
      .echo on
      .import {load_file} {table_name}
      .echo off
      -- @formatter:on
      """.rstrip()
    )

    sql_cmd_parts.append(sqlite_cmds)

    set_null_sql = prettify_sql("\n\n".join([col.create_update_blanks_to_null(table_name) for col in sqlite_cols if col.has_nulls]))
    sql_cmd_parts.append(set_null_sql)

  sql_cmd = "\n\n".join([part.strip() for part in sql_cmd_parts])

  result = SQLiteCreateTable(columns=sqlite_cols, sql_cmd=sql_cmd)

  if create_csv_from_rows:
    cols_fields = [col.name for col in sqlite_cols]
    if create_csv_convert_booleans:
      for row in rows:
        for col_name in cols_fields:
          value = row.get(col_name)
          if isinstance(value, bool):
            row[col_name] = str(value).upper()

    result.rows_as_csv = dicts_to_csv(
      rows,
      sep="|",
      header=False,
      fields_included=cols_fields,
      fields_order=cols_fields,
    )

    if load_file:
      result.csv_file = files.convert_to_path(load_file)
      files.write_file(result.csv_file, result.rows_as_csv, log_file_path=True)

  return result


def sqlite_vacuum(db_file: Path):
  from ltpylib.procs import run_with_regular_stdout

  prev_size = convert_decimal_precision(db_file.stat().st_size / 1024.0 / 1024.0, 2)
  sqlite_cmd = [
    "sqlite3",
    db_file.as_posix(),
    "VACUUM;",
  ]
  run_with_regular_stdout(sqlite_cmd, check=True, cwd=db_file.parent)
  new_size = convert_decimal_precision(db_file.stat().st_size / 1024.0 / 1024.0, 2)
  print(f"VACUUM result: db={db_file.name} before_mb={prev_size} after_mb={new_size}")



#!/usr/bin/env python
from typing import List, Sequence

from ltpylib.dicts import create_key_getter


def sort_by_key(vals: List[dict], key: str, reverse: bool = False) -> List[dict]:
  vals.sort(key=create_key_getter(key), reverse=reverse)
  return vals


def sort_by_keys(
  vals: List[dict],
  keys: Sequence[str],
  reverse: bool = False,
  skip_missing: bool = False,
) -> List[dict]:
  if skip_missing:

    def key_func(val: dict) -> list:
      return [val.get(key) for key in keys if key in val]

  else:

    def key_func(val: dict) -> list:
      return [val.get(key, "") for key in keys]

  vals.sort(key=key_func, reverse=reverse)
  return vals



#!/usr/bin/env python
# pylint: disable=C0111
import itertools
import json
import os
import re
import sys
from pathlib import Path
from subprocess import CalledProcessError
from typing import AnyStr, Callable, List, Match, Optional, Pattern, Sequence, Set, Tuple, Union

from ltpylib import gitrepos, inputs, logs, procs, strings
from ltpylib.common_types import TypeWithDictRepr
from ltpylib.macos import pbcopy


def convert_to_path(path: Union[Path, str]) -> Path:
  if isinstance(path, str):
    return Path(path)

  return path


def convert_to_path_expandvars(path: Union[Path, str]) -> Path:
  if isinstance(path, str):
    return Path(os.path.expandvars(path))

  return path


def replace_matches_in_file(
  file: Union[str, Path],
  search_string: str,
  replacement: Union[str, Callable[[Match], str]],
  quote_replacement: Union[bool, str] = False,
  wrap_replacement_in_function: Union[bool, str] = False,
  force_replace: bool = False,
  flags: Union[int, re.RegexFlag] = 0,
) -> bool:
  if isinstance(quote_replacement, str):
    quote_replacement = strings.convert_to_bool(quote_replacement)

  if isinstance(wrap_replacement_in_function, str):
    wrap_replacement_in_function = strings.convert_to_bool(wrap_replacement_in_function)

  if isinstance(force_replace, str):
    force_replace = strings.convert_to_bool(force_replace)

  if isinstance(flags, str):
    flags = strings.convert_to_number(flags)

  if quote_replacement and isinstance(replacement, str):
    replacement = re.escape(replacement)
  elif wrap_replacement_in_function and isinstance(replacement, str):
    replacement_content = replacement

    def replacement_function(match: Match) -> str:
      return replacement_content

    replacement = replacement_function

  content = read_file(file)
  content_new = re.sub(search_string, replacement, content, flags=flags)

  if content != content_new:
    write_file(file, content_new)
    return True
  elif force_replace and re.search(search_string, content, flags=flags):
    write_file(file, content_new)
    return True

  return False


def replace_strings_in_file(
  file: Union[str, Path],
  search_string: str,
  replacement: str,
  count: int = -1,
  force_replace: bool = False,
) -> bool:
  if isinstance(count, str):
    count = strings.convert_to_number(count)

  if isinstance(force_replace, str):
    force_replace = strings.convert_to_bool(force_replace)

  content: str = read_file(file)
  content_new = content.replace(search_string, replacement, count)

  if content != content_new:
    write_file(file, content_new)
    return True
  elif force_replace and search_string in content:
    write_file(file, content_new)
    return True

  return False


def remove_matching_lines_in_file(
  file: Union[str, Path],
  search_string: str,
  quote_search_string: bool = False,
  flags: Union[int, re.RegexFlag] = 0,
) -> bool:
  if isinstance(quote_search_string, str):
    quote_search_string = strings.convert_to_bool(quote_search_string)

  if isinstance(flags, str):
    flags = strings.convert_to_number(flags)

  if quote_search_string:
    search_string = re.escape(search_string)

  content: str = read_file(file)
  matcher = re.compile(search_string, flags=flags)
  has_match = False
  new_lines = []
  for line in content.splitlines():
    if matcher.search(line):
      has_match = True
    else:
      new_lines.append(line)

  if has_match:
    write_file(file, "\n".join(new_lines))
    return True

  return False


def chmod_proc(perms: str, file: Union[str, Path]) -> int:
  import subprocess

  file = convert_to_path(file)

  return subprocess.call(["chmod", perms, file.as_posix()])


def read_file(file: Union[str, Path]) -> str:
  file = convert_to_path(file)

  with open(file.as_posix(), 'r') as fr:
    content = fr.read()
  return content


def read_json_file(file: Union[str, Path]) -> Union[dict, list]:
  file = convert_to_path(file)

  with open(file.as_posix(), 'r') as fr:
    parsed = json.load(fr)

  return parsed


def read_yaml_file(file: Union[str, Path]) -> Union[dict, list]:
  import yaml

  file = convert_to_path(file)

  with open(file.as_posix(), 'r') as fr:
    parsed = yaml.full_load(fr)

  return parsed


def read_file_n_lines(file: Union[str, Path], n_lines: int = -1) -> List[str]:
  file = convert_to_path(file)

  lines: List[str] = []
  with open(file.as_posix()) as fr:
    if n_lines < 0:
      while True:
        line = fr.readline()
        if not line:
          break
        lines.append(line.rstrip('\n'))
    else:
      for n in range(n_lines):
        line = fr.readline()
        if not line:
          break
        lines.append(line.rstrip('\n'))

  return lines


def write_file(file: Union[str, Path], contents: AnyStr, log_file_path: bool = False):
  file = convert_to_path(file)

  with open(file.as_posix(), 'w') as fw:
    fw.write(contents)

  if log_file_path:
    import logging

    if not logging.root.hasHandlers():
      print(file.as_posix())
    else:
      logging.info(file.as_posix())


def write_file_and_log(file: Union[str, Path], contents: AnyStr):
  write_file(file, contents, log_file_path=True)


def append_file(file: Union[str, Path], contents: AnyStr):
  file = convert_to_path(file)

  with open(file.as_posix(), 'a') as fw:
    fw.write(contents)


def list_files(base_dir: Path, globs: Union[List[str], str] = ('**/*',)) -> List[Path]:
  if isinstance(globs, str):
    globs = [globs]

  files: Set[Path] = set()
  file: Optional[Path] = None
  for file in list(itertools.chain(*[base_dir.glob(glob) for glob in globs])):
    if file.is_file():
      files.add(file)

  files_list = list(files)
  files_list.sort()
  return files_list


def list_dirs(base_dir: Path, globs: List[str] = ('**/*',)) -> List[Path]:
  dirs: Set[Path] = set()
  child_dir: Optional[Path] = None
  for child_dir in list(itertools.chain(*[base_dir.glob(glob) for glob in globs])):
    if child_dir.is_dir():
      dirs.add(child_dir)

  dirs_list = list(dirs)
  dirs_list.sort()
  return dirs_list


def filter_files_with_matching_line(files: List[Union[str, Path]], regexes: List[Union[str, Pattern]], check_n_lines: int = 1) -> List[Path]:
  filtered: List[Path] = []
  for file in files:
    lines = read_file_n_lines(file, check_n_lines)
    has_match = False
    for line in lines:
      for regex in regexes:
        if re.search(regex, line):
          has_match = True
          break

      if has_match:
        break

    if not has_match:
      filtered.append(file)

  return filtered


def find_parent(
  base_dir: Union[Path, str],
  file_name: str,
  max_depth: int = -1,
) -> Optional[Path]:
  curr_dir = convert_to_path(base_dir)
  depth = max_depth

  while depth != 0:
    parent_file = curr_dir.joinpath(file_name)
    if parent_file.exists():
      return parent_file

    if curr_dir == curr_dir.parent:
      break

    curr_dir = curr_dir.parent
    depth -= 1

  return None


def find_children(
  base_dir: Union[Path, str],
  break_after_match: bool = False,
  max_depth: int = -1,
  include_dirs: bool = True,
  include_files: bool = True,
  match_absolute_path: bool = False,
  include_patterns: Sequence[str] = None,
  exclude_patterns: Sequence[str] = None,
  includes: Sequence[str] = None,
  excludes: Sequence[str] = None,
  recursion_include_patterns: Sequence[str] = None,
  recursion_exclude_patterns: Sequence[str] = None,
  recursion_includes: Sequence[str] = None,
  recursion_excludes: Sequence[str] = None
) -> List[Path]:
  if isinstance(base_dir, str):
    top = str(base_dir)
  else:
    top = base_dir.as_posix()

  found_dirs = []
  return _find_children(
    top,
    found_dirs,
    1,
    break_after_match,
    max_depth,
    include_dirs,
    include_files,
    match_absolute_path,
    include_patterns,
    exclude_patterns,
    includes,
    excludes,
    recursion_include_patterns,
    recursion_exclude_patterns,
    recursion_includes,
    recursion_excludes
  )[1]


def _find_children(
  top: str,
  found_dirs: List[Path],
  current_depth: int,
  break_after_match: bool,
  max_depth: int,
  include_dirs: bool,
  include_files: bool,
  match_absolute_path: bool,
  include_patterns: Sequence[str],
  exclude_patterns: Sequence[str],
  includes: Sequence[str],
  excludes: Sequence[str],
  recursion_include_patterns: Sequence[str],
  recursion_exclude_patterns: Sequence[str],
  recursion_includes: Sequence[str],
  recursion_excludes: Sequence[str]
) -> Tuple[bool, List[Path]]:
  from ltpylib import filters

  found_match = False
  scandir_it = os.scandir(top)
  dirs = []

  with scandir_it:
    while True:
      try:
        try:
          entry = next(scandir_it)
        except StopIteration:
          break
      except OSError:
        return found_match, found_dirs

      try:
        is_dir = entry.is_dir()
      except OSError:
        # If is_dir() raises an OSError, consider that the entry is not
        # a directory, same behaviour than os.path.isdir().
        is_dir = False

      if is_dir and not include_dirs:
        continue
      elif not is_dir and not include_files:
        continue

      child = entry.name
      full_path = os.path.join(top, child)
      test_value = child if not match_absolute_path else full_path
      include = filters.should_include(
        test_value,
        include_patterns=include_patterns,
        exclude_patterns=exclude_patterns,
        includes=includes,
        excludes=excludes,
      )

      if include:
        found_match = True
        found_dirs.append(Path(full_path))
        if break_after_match:
          break

      if is_dir:
        include_child = filters.should_include(
          test_value,
          include_patterns=recursion_include_patterns,
          exclude_patterns=recursion_exclude_patterns,
          includes=recursion_includes,
          excludes=recursion_excludes,
        )
        if include_child:
          dirs.append(child)

  if (max_depth <= -1 or current_depth < max_depth) and (not found_match or not break_after_match):
    for dirname in dirs:
      _find_children(
        os.path.join(top, dirname),
        found_dirs,
        current_depth + 1,
        break_after_match,
        max_depth,
        include_dirs,
        include_files,
        match_absolute_path,
        include_patterns,
        exclude_patterns,
        includes,
        excludes,
        recursion_include_patterns,
        recursion_exclude_patterns,
        recursion_includes,
        recursion_excludes
      )

  return found_match, found_dirs


class SplitLine(TypeWithDictRepr):

  def __init__(self, file_name: str, line_number: int = None, content: str = None):
    self.file_name: str = file_name
    self.line_number: int = line_number
    self.content: str = content

  def to_file_name_and_content(self) -> str:
    return ":".join([val for val in [self.file_name, self.content] if val is not None])

  def to_open_arg(self) -> str:
    if self.line_number is not None:
      return "%s:%s" % (self.file_name, self.line_number)

    return self.file_name


class OpenGreppedLines:
  CHOICE_PRINT = "Print"
  CHOICE_COPY = "Copy"
  CHOICE_IDEA = "Open in IntelliJ"
  CHOICE_CODE = "Open in VSCode"
  APP_IDEA = "idea"
  APP_CODE = "code"

  REPLACE_HOME_REGEX = re.compile(r"(?m)^~")
  LINE_NUMBERS_COLON_STRICT_REGEX = re.compile(r"^(\S+):(\d+):(.*)")
  LINE_NUMBERS_COLON_REGEX = re.compile(r"^(.*\S):(\d+):(.*)")
  LINE_NUMBERS_DASH_STRICT_REGEX = re.compile(r"^(\S+)-(\d+)-(.*)")
  LINE_NUMBERS_DASH_REGEX = re.compile(r"^(.*\S)-(\d+)-(.*)")
  LINE_NUMBERS_EXTRACTOR_REGEX = re.compile(r"^(.*?\S)[:-](\d+)[:-](.*)")
  CHECK_REGEXES = [LINE_NUMBERS_COLON_STRICT_REGEX, LINE_NUMBERS_DASH_STRICT_REGEX, LINE_NUMBERS_COLON_REGEX, LINE_NUMBERS_DASH_REGEX]

  def __init__(
    self,
    results: str,
    ask: bool = None,
    line_numbers: bool = None,
    lines_exit_code: int = None,
    open_in_app: str = None,
    select_header: str = None,
    select_lines: bool = None,
    select_preview_file_cmd: str = None,
    skip_code: bool = None,
    skip_copy: bool = None,
    skip_idea: bool = None,
    skip_print: bool = None,
    debug: bool = None,
  ):
    self.ask: bool = ask
    self.line_numbers: bool = line_numbers
    self.lines_exit_code: int = lines_exit_code
    self.open_in_app: str = open_in_app
    self.select_header: str = select_header
    self.select_lines: bool = select_lines
    self.select_preview_file_cmd: str = select_preview_file_cmd
    self.skip_code: bool = skip_code
    self.skip_copy: bool = skip_copy
    self.skip_idea: bool = skip_idea
    self.skip_print: bool = skip_print
    self.debug: bool = debug

    self.results: Optional[str] = None
    if not results or results[0] == "-":
      self.using_stdin: bool = True
      if not self.select_lines:
        self.results: str = sys.stdin.read()
    else:
      self.using_stdin: bool = False
      self.results: str = results

    if self.results:
      self.results: str = self.results.strip()

    self.original_results: str = self.results
    self.check_for_line_numbers()

  def check_for_line_numbers(self) -> bool:
    if self.line_numbers is None and self.results:
      if ":" in self.results or OpenGreppedLines.LINE_NUMBERS_DASH_REGEX.match(self.results):
        self.line_numbers = True
      else:
        self.line_numbers = False

    return self.line_numbers

  def main(self):

    if self.lines_exit_code != 0:
      print(self.results)
      exit(self.lines_exit_code)

    if self.select_lines:
      self.handle_select_lines()

    if not self.results.strip():
      exit(0)

    if self.ask:
      selected_action: str = self.ask_for_action()
    else:
      selected_action: str = self.open_in_app

    if selected_action == OpenGreppedLines.CHOICE_PRINT:
      self.handle_print()
    elif selected_action == OpenGreppedLines.CHOICE_COPY:
      self.handle_copy()
    elif selected_action in [OpenGreppedLines.CHOICE_IDEA, OpenGreppedLines.APP_IDEA]:
      self.handle_open_in_app([OpenGreppedLines.APP_IDEA])
    elif selected_action in [OpenGreppedLines.CHOICE_CODE, OpenGreppedLines.APP_CODE]:
      self.handle_open_in_app([OpenGreppedLines.APP_CODE, "-r", "-g"])
    elif self.open_in_app and self.open_in_app != OpenGreppedLines.APP_IDEA:
      self.handle_open_in_app([self.open_in_app])
    else:
      print("FATAL action not recognized: %s" % selected_action)
      exit(1)

  def handle_copy(self):
    pbcopy(self.get_cleaned_results().strip())

  def handle_print(self):
    print(self.results)

  def handle_open_in_app(self, open_args: List[str]):
    for open_file in self.get_files_from_results():
      proc_args = open_args + [open_file]
      if self.debug:
        print(proc_args)
      else:
        procs.run_with_regular_stdout(proc_args)

  def create_select_lines_preview_command(self) -> str:
    if self.line_numbers:
      filename_regex = OpenGreppedLines.LINE_NUMBERS_EXTRACTOR_REGEX.pattern
      lineno_cmd = f"$(echo {{}} | pcregrep -o2 '{filename_regex}')"
    else:
      filename_regex = r"^([^:]+)"
      lineno_cmd = ""

    return f"""
filename="$(echo {{}} | pcregrep -o1 '{filename_regex}')"
filepath="$(echo "$filename" | sed 's|~|{Path.home().as_posix()}|')"
lineno="{lineno_cmd}"
if test -n "$filename" && test -e "$filepath"; then
  if test -n "$lineno"; then
    echo "${{filename}}:${{lineno}}"
  else
    echo "${{filename}}"
  fi
  echo "{logs.LOG_SEP}"
  {self.select_preview_file_cmd}
else
  echo "No file found to display: filename=$filename filepath=$filepath lineno=$lineno"
fi
"""

  def handle_select_lines(self):
    bind_open_cmd = "open_grepped_lines"
    if self.line_numbers is not None:
      bind_open_cmd += " --line-numbers" if self.line_numbers else " --no-line-numbers"

    self.results = inputs.select_prompt(
      None if self.using_stdin else self.original_results.splitlines(),
      stdin=sys.stdin if self.using_stdin else None,
      header=self.select_header,
      multi=True,
      ansi=True,
      preview=self.create_select_lines_preview_command(),
      binds=[
        "?:toggle-preview",
        f"right:execute-silent({bind_open_cmd} {{}})",
        "ctrl-a:select-all",
      ],
    )
    self.check_for_line_numbers()

  def split_line(self, line: str) -> SplitLine:
    if not self.line_numbers:
      return SplitLine(line)

    for regex in OpenGreppedLines.CHECK_REGEXES:
      match = regex.fullmatch(line)
      if match:
        return SplitLine(match.group(1), line_number=int(match.group(2)), content=match.group(3))

    return SplitLine(line)

  def get_files_from_results(self) -> List[str]:
    results = self.get_cleaned_results()
    files_from_results: List[str] = []
    found_files = []

    for line in results.splitlines():
      line_parts = self.split_line(line)
      if line_parts.file_name in found_files:
        continue

      found_files.append(line_parts.file_name)

      file_path = Path(line_parts.file_name)
      if not file_path.exists():
        file_path = gitrepos.resolve_file_relative_to_git_base_dir(file_path)
        if file_path is not None and file_path.exists():
          line_parts.file_name = file_path.as_posix()
        else:
          raise ValueError(f"could not parse file from results: {line_parts.file_name}")

      files_from_results.append(line_parts.to_open_arg())

    return files_from_results

  def ask_for_action(self) -> str:
    try:
      action = inputs.select_prompt(
        self.get_action_choices(),
        header="Action for lines",
      )

      return action
    except CalledProcessError:
      print(self.results)
      exit(0)

  def get_action_choices(self) -> List[str]:
    choices: List[str] = []

    if not self.skip_idea:
      choices.append(OpenGreppedLines.CHOICE_IDEA)

    if not self.skip_code:
      choices.append(OpenGreppedLines.CHOICE_CODE)

    if not self.skip_copy:
      choices.append(OpenGreppedLines.CHOICE_COPY)

    if not self.skip_print:
      choices.append(OpenGreppedLines.CHOICE_PRINT)

    return choices

  def get_cleaned_results(self) -> str:
    return OpenGreppedLines.REPLACE_HOME_REGEX.sub(
      Path.home().as_posix(),
      strings.strip_color_codes(self.results.strip()),
    )


def _main():
  import sys

  result = globals()[sys.argv[1]](*sys.argv[2:])
  if result is not None:
    print(result)


if __name__ == "__main__":
  try:
    _main()
  except KeyboardInterrupt:
    exit(130)



#!/usr/bin/env python
import re
from decimal import Decimal
from typing import List, Match, Optional, Union

BOOLEAN_STRINGS_FALSE = frozenset([
  "no",
  "n",
  "false",
])
BOOLEAN_STRINGS_TRUE = frozenset([
  "yes",
  "y",
  "true",
])

CAMEL_CASE_CAP_CHARS_REGEX = re.compile(r"(?<=[a-z])([A-Z0-9])|(?<=[^0-9])([A-Z])(?=[a-z])")
CASE_CONVERSION_IGNORE_REGEX = re.compile(r"[']")
MULTI_SPACE_REGEX = re.compile(r"\s+")
NON_ALPHA_NUMERIC_REGEX = re.compile(r"[^a-zA-Z0-9]")
HUMAN_SIZE_REGEX = re.compile(r"^([\d.]+)([a-zA-Z]{,2})$")
HUMAN_SIZE_UNITS_LOOKUP = {
  "tb": 1_000_000_000_000,
  "gb": 1_000_000_000,
  "mb": 1_000_000,
  "kb": 1_000,
  "b": 1,
}


def camel_case_to_title_case(val: str):
  val = val[:1].upper() + val[1:]
  return " ".join(re.findall(r'[A-Z](?:[a-z]+|[A-Z]*(?=[A-Z]|$))', val))


def convert_to_bool(
  val: str,
  check_if_valid: bool = False,
  unknown_value: bool = None,
  none_value: Union[bool, None] = None,
) -> Union[bool, str, None]:
  if val is None:
    return none_value

  if check_if_valid and not is_boolean(val):
    return val

  lower_val = val.lower()

  if lower_val in BOOLEAN_STRINGS_FALSE:
    return False
  elif lower_val in BOOLEAN_STRINGS_TRUE:
    return True
  elif unknown_value is not None:
    return unknown_value

  raise ValueError("String is not a boolean: " % val)


def convert_to_number(
  val: str,
  check_if_valid: bool = False,
  float_only: bool = False,
  use_decimal: bool = False,
  remove_commas: bool = False,
) -> Union[int, float, str, None]:
  if val is None:
    return None

  if check_if_valid and not is_number(val, allow_comma=remove_commas):
    return val

  if remove_commas:
    val = val.replace(",", "")

  if float_only:
    return Decimal(val) if use_decimal else float(val)

  try:
    return int(val)
  except ValueError:
    return Decimal(val) if use_decimal else float(val)


def convert_human_size_to_number(val: str, in_unit: str = None) -> Optional[float]:
  match = HUMAN_SIZE_REGEX.match(val)
  if match:
    num = float(match.group(1))
    units = match.group(2)
    multiplier = HUMAN_SIZE_UNITS_LOOKUP.get(units.lower())
    if not multiplier:
      raise ValueError("missing unit: " + units)

    result = num * multiplier
    if in_unit:
      divisor = HUMAN_SIZE_UNITS_LOOKUP.get(in_unit.lower())
      if not divisor:
        raise ValueError("missing unit: " + in_unit)

      result = result / divisor

    return result

  return None


def is_boolean(val: str) -> bool:
  if not val:
    return False

  return val.lower() in BOOLEAN_STRINGS_FALSE or val.lower() in BOOLEAN_STRINGS_TRUE


def is_number(val: str, allow_comma: bool = False) -> bool:
  if not val:
    return False

  if val.isdigit():
    return True

  val = val.replace(".", "", 1)
  if val.isdigit():
    return True

  if val.startswith("-") and val.replace("-", "", 1).isdigit():
    return True

  if allow_comma and val.replace(",", "").isdigit():
    return True

  return False


def maybe_json_string(val: str) -> bool:
  if not val:
    return False

  if val.startswith("[") and val.endswith("]"):
    return True

  return val.startswith("{") and val.endswith("}")


def str_list_max_length(values: List[str]) -> int:
  return len(max(values, key=len))


def strip_color_codes(val: str) -> str:
  return re.sub(r"\x1B\[([0-9]{1,2}(;[0-9]{1,2})?)?[mGK]", "", val)


def substring_after(val: str, before_str: str) -> str:
  return val.split(before_str, 1)[1]


def substring_after_last(val: str, sep: str) -> str:
  return val.split(sep)[-1]


def substring_before(val: str, before_str: str) -> str:
  return val.split(before_str)[0]


def substring_before_last(val: str, before_str: str) -> str:
  parts = val.split(before_str)
  if len(parts) == 1:
    return parts[0]

  return before_str.join(parts[0:-1])


def _to_snake_case_replacer(match: Match) -> str:
  return " " + "".join([val for val in match.group(1, 2) if val is not None])


def to_snake_case(val: str) -> str:
  val = CASE_CONVERSION_IGNORE_REGEX.sub("", val)

  val = NON_ALPHA_NUMERIC_REGEX.sub(" ", val)

  val = CAMEL_CASE_CAP_CHARS_REGEX.sub(_to_snake_case_replacer, val)

  return MULTI_SPACE_REGEX.sub("_", val.lower().strip())


def truncate_if_needed(val: str, max_length: int, include_ellipsis: bool = True) -> str:
  if val and len(val) > max_length:
    return (val[:max_length] + "..") if include_ellipsis else val[:max_length]

  return val


def _main():
  import sys

  result = globals()[sys.argv[1]](*sys.argv[2:])
  if result is not None:
    print(result)


if __name__ == "__main__":
  try:
    _main()
  except KeyboardInterrupt:
    exit(130)



#!/usr/bin/env python
import unittest

from ltpylib.common_types import DataWithUnknownProperties, DataWithUnknownPropertiesAsAttributes


class TestDataWithUnknownProperties(unittest.TestCase):

  def test_DataWithUnknownProperties_empty(self):
    val = DataWithUnknownProperties(values={})
    self.assertIsNone(val.unknown_properties)

  def test_DataWithUnknownProperties_none(self):
    val = DataWithUnknownProperties(values=None)
    self.assertIsNone(val.unknown_properties)

  def test_DataWithUnknownProperties_some(self):
    values = {
      "field_num": 3,
      "field_str": "abc",
    }
    val = DataWithUnknownProperties(values=values)
    self.assertDictEqual(val.unknown_properties, values)


class TestDataWithUnknownPropertiesAsAttributes(unittest.TestCase):

  def test_DataWithUnknownPropertiesAsAttributes(self):
    values = {
      "field_num": 3,
      "field_str": "abc",
    }
    val = DataWithUnknownPropertiesAsAttributes(values=values)

    self.assertTrue(val.has_unknown_properties)
    self.assertFalse(hasattr(val, "unknown_properties"))

    self.assertTrue(hasattr(val, "field_num"))
    self.assertEqual(getattr(val, "field_num"), 3)
    self.assertEqual(val.field_num, 3)

    self.assertTrue(hasattr(val, "field_str"))
    self.assertEqual(getattr(val, "field_str"), "abc")
    self.assertEqual(val.field_str, "abc")

  def test_DataWithUnknownPropertiesAsAttributes_none(self):
    val = DataWithUnknownPropertiesAsAttributes(values=None)
    self.assertFalse(val.has_unknown_properties)
    self.assertFalse(hasattr(val, "unknown_properties"))

    val = DataWithUnknownPropertiesAsAttributes(values={})
    self.assertFalse(val.has_unknown_properties)
    self.assertFalse(hasattr(val, "unknown_properties"))


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
from datetime import datetime, tzinfo
from typing import Optional

import pytz


def add_timezone_if_missing(date: datetime, tz: tzinfo) -> datetime:
  if date.tzinfo is not None:
    return date

  return date.astimezone(tz)


def as_local_timezone(date: datetime) -> datetime:
  from dateutil import tz

  return date.astimezone(tz.tzlocal())


def as_pacific_time(date: datetime) -> datetime:
  return date.astimezone(pytz.timezone("US/Pacific"))


def from_millis(millis: int) -> datetime:
  return datetime.fromtimestamp(millis / 1000.0)


def is_last_day_of_month(date: datetime) -> datetime:
  import calendar

  return calendar.monthrange(date.year, date.month)[1] == date.day


def to_last_day_of_month(date: datetime) -> datetime:
  import calendar

  return date.replace(day=calendar.monthrange(date.year, date.month)[1])


def parse_iso_date(date_string: str) -> Optional[datetime]:
  if not date_string:
    return None

  from dateutil import parser

  return parser.isoparse(date_string)


def parse_date(date_string: str, date_format: str = None) -> Optional[datetime]:
  if not date_string:
    return None

  if isinstance(date_string, datetime):
    return date_string

  from dateutil import parser

  if not date_format:
    return parser.parse(date_string)

  return datetime.strptime(date_string, date_format)


def parse_possibly_relative_date(date_string: str) -> datetime:
  import dateparser

  return dateparser.parse(date_string)


def to_yyyymmdd(date: datetime) -> str:
  return date.strftime("%Y%m%d")


def to_yyyymmdd_dashes(date: datetime) -> str:
  return date.strftime("%Y-%m-%d")


def to_json_isoformat(date: datetime) -> str:
  return date.isoformat(sep="T", timespec="milliseconds") + "Z"


def to_json_isoformat_friendly(date: datetime) -> str:
  return date.isoformat(sep=" ", timespec="auto")


def to_hhMMss_ampm(date: datetime) -> str:
  return date.strftime("%I:%M:%S %p")


def add(
  date: datetime,
  years: int = 0,
  months: int = 0,
  days: int = 0,
  leapdays: int = 0,
  weeks: int = 0,
  hours: int = 0,
  minutes: int = 0,
  seconds: int = 0,
  microseconds: int = 0,
  year: int = None,
  month: int = None,
  day: int = None,
  weekday: int = None,
  yearday: int = None,
  nlyearday: int = None,
  hour: int = None,
  minute: int = None,
  second: int = None,
  microsecond: int = None,
) -> datetime:
  from dateutil.relativedelta import relativedelta

  return date + relativedelta(
    years=years,
    months=months,
    days=days,
    leapdays=leapdays,
    weeks=weeks,
    hours=hours,
    minutes=minutes,
    seconds=seconds,
    microseconds=microseconds,
    year=year,
    month=month,
    day=day,
    weekday=weekday,
    yearday=yearday,
    nlyearday=nlyearday,
    hour=hour,
    minute=minute,
    second=second,
    microsecond=microsecond,
  )


def _main():
  import sys

  result = globals()[sys.argv[1]](*sys.argv[2:])
  if result is not None:
    print(result)


if __name__ == "__main__":
  try:
    _main()
  except KeyboardInterrupt:
    exit(130)



#!/usr/bin/env python
import logging
import os
import subprocess
import sys
from typing import IO, List, Optional, Sequence

from ltpylib import procs, strings
from ltpylib.strings import strip_color_codes


def confirm(question: str = 'Continue?', default: str = None) -> bool:
  """Ask a yes/no question via raw_input() and return their answer.

  "question" is a string that is presented to the user.
  "default" is the presumed answer if the user just hits <Enter>.
      It must be "yes" (the default), "no" or None (meaning
      an answer is required of the user).

  The "answer" return value is True for "yes" or False for "no".
  """
  valid = {
    "yes": True,
    "y": True,
    "ye": True,
    "no": False,
    "n": False,
  }
  if default is None:
    prompt = "\n[y/n]> "
  elif default == "yes" or default == "y":
    prompt = "\n[Y/n]> "
  elif default == "no" or default == "n":
    prompt = "\n[y/N]> "
  else:
    raise ValueError("invalid default answer: '%s'" % default)

  while True:
    choice = input(question + prompt).lower()
    if default is not None and choice == '':
      return valid[default]
    elif choice in valid:
      return valid[choice]
    else:
      logging.error("Please respond with 'yes' or 'no' (or 'y' or 'n').\n")


def confirm_default_yes(question: str = 'Continue?') -> bool:
  return confirm(question=question, default="y")


def confirm_with_auto(question: str = 'Continue?', auto_confirm_arg: bool = None, default: str = "y") -> bool:
  if check_auto_confirm(auto_confirm_arg):
    return True

  return confirm(question=question, default=default)


def check_auto_confirm(auto_confirm_arg: bool = None) -> bool:
  if auto_confirm_arg is True:
    return True

  return auto_confirm_from_env_var()


def auto_confirm_from_env_var() -> bool:
  auto_confirm = os.getenv("auto_confirm")
  return auto_confirm is not None and strings.is_boolean(auto_confirm) and strings.convert_to_bool(auto_confirm)


def ask_for_input(prompt: str = 'Please input value.') -> str:
  return input(prompt + '\n> ')


def create_fzf_select_prompt_command(
  header: str = None,
  no_sort: bool = True,
  layout: str = "reverse",
  multi: bool = False,
  ansi: bool = False,
  preview: str = None,
  preview_window: str = "down:wrap:hidden",
  binds: Sequence[str] = None,
  include_select_all_bind: bool = False,
  fzf_args: Sequence[str] = None,
) -> List[str]:
  command = [
    "fzf",
    "--layout",
    layout,
  ]

  if header:
    command.extend([
      "--header",
      header,
    ])

  if no_sort:
    command.append("--no-sort")

  if multi:
    command.append("--multi")

  if ansi:
    command.append("--ansi")

  if preview:
    command.extend([
      "--preview",
      preview,
      "--preview-window",
      preview_window,
    ])

  if include_select_all_bind:
    if not binds:
      binds = []
    else:
      binds = list(binds)

    binds.append("ctrl-a:select-all")

  if binds:
    for bind in binds:
      command.extend([
        "--bind",
        bind,
      ])

  if fzf_args:
    command.extend(fzf_args)

  return command


def select_prompt(
  choices: Optional[Sequence[str]],
  stdin: Optional[IO] = None,
  header: str = None,
  no_sort: bool = True,
  layout: str = "reverse",
  multi: bool = False,
  ansi: bool = False,
  preview: str = None,
  preview_window: str = "down:wrap:hidden",
  binds: Sequence[str] = None,
  include_select_all_bind: bool = False,
  fzf_args: Sequence[str] = None,
) -> str:
  command = create_fzf_select_prompt_command(
    header=header,
    no_sort=no_sort,
    layout=layout,
    multi=multi,
    ansi=ansi,
    preview=preview,
    preview_window=preview_window,
    binds=binds,
    include_select_all_bind=include_select_all_bind,
    fzf_args=fzf_args,
  )

  result: subprocess.CompletedProcess = procs.run(
    command,
    universal_newlines=True,
    stderr=sys.stderr,
    input=None if stdin is not None else "\n".join(choices),
    stdin=stdin,
  )

  if result.returncode == 130:
    raise KeyboardInterrupt

  if result.returncode == 1:
    exit(0)

  result.check_returncode()
  return result.stdout.strip()


def select_prompt_and_return_indexes(
  choices: Sequence[str],
  header: str = None,
  multi: bool = False,
  ansi: bool = False,
  preview: str = None,
  preview_window: str = "down:wrap:hidden",
  binds: Sequence[str] = None,
  include_select_all_bind: bool = False,
  fzf_args: Sequence[str] = None,
) -> List[int]:
  selections = select_prompt(
    choices,
    header=header,
    no_sort=True,
    layout="reverse",
    multi=multi,
    ansi=ansi,
    preview=preview,
    preview_window=preview_window,
    binds=binds,
    include_select_all_bind=include_select_all_bind,
    fzf_args=fzf_args,
  ).splitlines(keepends=False)

  choices_without_colors = [strip_color_codes(choice) for choice in choices]

  return [choices_without_colors.index(item) for item in selections]


def select_prompt_cmd(
  choices: Sequence[str],
  message: str = None,
  bottom_message: str = None,
  max_width: int = None,
) -> str:
  command = ["select_prompt"]

  if message:
    command.extend(["--message", message])

  if bottom_message:
    command.extend(["--bottom-message", message])

  if max_width:
    command.extend(["--max-width", str(max_width)])

  command.extend(choices)

  result: subprocess.CompletedProcess = procs.run(
    command,
    universal_newlines=True,
    stderr=sys.stderr,
    stdin=sys.stdin,
  )
  if result.returncode == 130:
    raise KeyboardInterrupt

  result.check_returncode()
  return result.stdout.strip()


def _main():
  import sys

  result = globals()[sys.argv[1]](*sys.argv[2:])
  if result is not None:
    print(result)


if __name__ == "__main__":
  try:
    _main()
  except KeyboardInterrupt:
    exit(130)
